/*
 * Wilcard X100P FXO Interface Driver for Zapata Telephony interface
 *
 * Ported on FreeBSD by
 *      Alexander Timoshenko <gonzo@portaone.com>, PortaOne software
 *
 * Based on previous works, designs, and architectures conceived written 
 * by 
 *							 Mark Spencer <markster@linux-support.net>
 *							 Matthew Fredrickson <creslin@linux-support.net>
 *
 * Copyright (C) 2001, Linux Support Services, Inc.
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */



/* $Id: wcfxo.c 101 2005-03-07 21:36:56Z riz $ */

#include <sys/param.h>  /* defines used in kernel.h */
#include <sys/module.h>
#include <sys/systm.h>
#include <sys/errno.h>
#include <sys/kernel.h> /* types used in module initialization */
#include <sys/conf.h>   /* cdevsw struct */
#include <sys/uio.h>	/* uio struct */
#include <sys/malloc.h>
#include <sys/lock.h>
#include <sys/endian.h>
#include <sys/bus.h>  /* structs, prototypes for pci bus stuff */

#include <machine/bus.h>
#include <sys/rman.h>
#include <machine/resource.h>
#include <machine/clock.h>
#include <sys/sysctl.h>

#include <vm/vm.h>		/* vtophys */
#include <vm/pmap.h>	/* vtophys */

#if __FreeBSD_version < 500000
#include <pci/pcireg.h>
#include <pci/pcivar.h>
#else
#include <dev/pci/pcireg.h>
#include <dev/pci/pcivar.h>
#endif


#include "zaptel.h"
#include "wcfxovar.h"
#include "wcfxoreg.h"

#include <ddb/ddb.h>


#define DEBUG 1
#define ZERO_BATT_RING 1
#define JAPAN 1
#define WC_MAX_IFACES 128
#define WCFXO_MAX_PORTS 4

#define RING_DEBOUNCE   64			  /* Ringer Debounce (in ms) */
#ifdef  JAPAN
#define BATT_DEBOUNCE   30			  /* Battery debounce (in ms) */
#define OH_DEBOUNCE	 350			 /* Off/On hook debounce (in ms) */
#else
#define BATT_DEBOUNCE   80			  /* Battery debounce (in ms) */
#endif


#define MINPEGTIME	  10 * 8		  /* 30 ms peak to peak gets us no more than 100 Hz */
#define PEGTIME		 50 * 8		  /* 50ms peak to peak gets us rings of 10 Hz or more */
#define PEGCOUNT		5			   /* 5 cycles of pegging means RING */


#define WC_CNTL	 0x00
#define WC_OPER	 0x01
#define WC_AUXC	 0x02
#define WC_AUXD	 0x03
#define WC_MASK0	0x04
#define WC_MASK1	0x05
#define WC_INTSTAT  0x06

#define WC_DMAWS	0x08
#define WC_DMAWI	0x0c
#define WC_DMAWE	0x10
#define WC_DMARS	0x18
#define WC_DMARI	0x1c
#define WC_DMARE	0x20

#define WC_AUXFUNC  0x2b
#define WC_SERCTL   0x2d
#define WC_FSCDELAY 0x2f

#define FLAG_EMPTY  0
#define FLAG_WRITE  1
#define FLAG_READ   2

#define FLAG_INVERTSER	  (1 << 0)
#define FLAG_USE_XTAL	   (1 << 1)
#define FLAG_DOUBLE_CLOCK   (1 << 2)
#define FLAG_RESET_ON_AUX5  (1 << 3)

#ifndef PCIR_BARS
#define PCIR_BARS		0x10
#define PCIR_BAR(x)		(PCIR_BARS + (x) * 4)
#endif

#ifndef MIN
#define MIN(a,b) (((a)<(b))?(a):(b))
#endif


static char * card_configuration = 0;

/* devices list */
static struct wcfxo_desc wcx100p = { "Wildcard X100P",
	FLAG_INVERTSER | FLAG_USE_XTAL | FLAG_DOUBLE_CLOCK };

static struct wcfxo_desc wcx101p = { "Wildcard X101P",
	FLAG_USE_XTAL | FLAG_DOUBLE_CLOCK };

static struct wcfxo_desc generic = { "Generic Clone",
	FLAG_USE_XTAL | FLAG_DOUBLE_CLOCK };


static struct wcfxo_ident wcfxo_ident_table[] = {
	/* {vendor, device, subverndor, card info }, */
	{ 0xe159, 0x0001, 0x8085,	&wcx101p },
	{ 0xe159, 0x0001, 0x8086, 	&generic },
	{ 0x1057, 0x5608, -1, 		&wcx100p },
	{ 0, 0, 0, 0},
};

static int wecareregs[] =
{ 5, 6, 9, 11, 12, 13, 17, 19, };


struct reg {
	unsigned long flags;
	unsigned char index;
	unsigned char reg;
	unsigned char value;
};


static struct fxo_mode {
	char *name;
	int ohs;
	int act;
	int dct;
	int rz;
	int rt;
	int lim;
	int vol;
} fxo_modes[] =
{
	{ "FCC", 0, 0, 2, 0, 0, 0, 0 },	 /* US */
	{ "CTR21", 0, 0, 3, 0, 0, 3, 0 },   /* Austria, Belgium, Denmark, Finland, France, Germany, 
										   Greece, Iceland, Ireland, Italy, Luxembourg, Netherlands,
										   Norway, Portugal, Spain, Sweden, Switzerland, and UK */
	{ "FCC", 0, 0, 1, 0, 0, 2, 0 } /* Russia */
};


struct wcfxo_softc {
	const struct 		wcfxo_device_description * sc_desc;
	/* card-global dynamic data */
	int					barmuxed;

	struct resource		* wc_irq;
	void *				irq_handler;
	bus_space_tag_t		wc_irq_bt;
	bus_space_handle_t	wc_irq_bh;

	void				*intr_cookie;
	int					ilr_enabled;

#if __FreeBSD_version < 500000
	dev_t				d;
#else
	struct cdev *		d;
#endif
	char 				*variety;
	struct zt_span 		span;
	struct zt_chan 		chan;
	int 				usecount;
	int 				dead;
	int 				pos;
	unsigned long 		flags;
	int 				freeregion;
	int 				ring;
	int 				offhook;
	int 				battery;
	int 				wregcount;
	int 				readpos;
	int 				rreadpos;
	unsigned int 		pegtimer;
	int 				pegcount;
	int 				peg;
	int 				battdebounce;
	int 				nobatttimer;
	int 				ringdebounce;
#ifdef  JAPAN
	int 				ohdebounce;
#endif
	int 				allread;
	/* How far off our registers are from what we expect */
	int 				regoffset;		  
	int 				alt;
	int 				ignoreread;
	int 				reset;
	/* Up to 6 register can be written at a time */
	struct reg 			regs[ZT_CHUNKSIZE];
	struct reg 			oldregs[ZT_CHUNKSIZE];
	unsigned char 		lasttx[ZT_CHUNKSIZE];
	/* Up to 32 registers of whatever we most recently read */
	unsigned char 		readregs[32];
	struct resource *	wc_base_addr;
	bus_space_tag_t 	wc_base_addr_tag;
	bus_space_handle_t 	wc_base_addr_handle;
	int					wc_ioport_rid;

	struct wcfxo_data	wc_ldata;
#ifdef ZERO_BATT_RING
	int 				onhook;
#endif
};

/* Function prototypes */
int 		wcfxo_hardware_init(struct wcfxo_softc *);
void 		wcfxo_enable_interrupts(struct wcfxo_softc *);
void 		wcfxo_start_dma(struct wcfxo_softc *);
static void wcfxo_dma_map_addr(void *, bus_dma_segment_t *, int, int);
static int	wcfxo_detach(device_t );
static void wcfxo_sysctl_init(void);
static void wcfxo_sysctl_destroy(void);
void		wcfxo_generate_description(struct wcfxo_softc *sc);

/* sysctl stuff */
extern struct sysctl_oid * zt_root;
static struct sysctl_oid * wcfxo_root;
static struct sysctl_ctx_list wcfxo_clist;

/* syscall handlers */
d_open_t wcfxo_sys_open;
d_close_t wcfxo_sys_close;
d_read_t wcfxo_sys_read;



/* Character device entry points */
static struct cdevsw wcfxo_cdevsw = {
#if __FreeBSD_version >= 502103
	.d_version = D_VERSION,
#endif
	.d_name =   "wcfxo",
	.d_open  =   wcfxo_sys_open,
	.d_close =   wcfxo_sys_close,
	.d_read  =   wcfxo_sys_read
};

static int debug = 0;
static int monitor = 0;
static int quiet = 0;
static int boost = 0;
static int opermode = 2;
static int batterycheck = 1;

static void wcfxo_release(struct wcfxo_softc *sc)
{
	zt_unregister(&sc->span);
	printf("Freed a Wildcard\n");
}

void wcfxo_start_dma(struct wcfxo_softc *sc)
{
	/* Reset Master and TDM */
	WRITE1(WC_CNTL, 0x0f);
	tsleep(sc, PZERO | PCATCH, "startdma", 1);
	/* Make sure that it is read */
	WRITE1(WC_CNTL, 0x01);
	WRITE1(WC_OPER, 0x01);
}

static void wcfxo_restart_dma(struct wcfxo_softc *sc)
{
	/* Reset Master and TDM */
	WRITE1(WC_CNTL, 0x01);
	WRITE1(WC_OPER, 0x01);
}


static void wcfxo_stop_dma(struct wcfxo_softc *sc)
{
	WRITE1(WC_OPER, 0x00);
}

static void wcfxo_reset_tdm(struct wcfxo_softc *sc)
{
	/* Reset TDM */
	WRITE1(WC_CNTL, 0x0f);
}

static void wcfxo_disable_interrupts(struct wcfxo_softc *sc)
{
	WRITE1(WC_MASK0, 0x00);
	WRITE1(WC_MASK1, 0x00);
}

static int wcfxo_setreg(struct wcfxo_softc *sc, unsigned char reg, unsigned char value)
{
	int x;
	if (sc->wregcount < ZT_CHUNKSIZE) {
		x = sc->wregcount;
		sc->regs[x].reg = reg;
		sc->regs[x].value = value;
		sc->regs[x].flags = FLAG_WRITE;
		sc->wregcount++;
		return 0;
	}
	printf("wcfxo: Out of space to write register %02x with %02x\n", reg, value);
	return -1;
}

void wcfxo_generate_description(struct wcfxo_softc *sc)
{
	if(!card_configuration)
		card_configuration = malloc(80 + 40 +  25 + 20, M_DEVBUF, M_NOWAIT);
	snprintf(card_configuration, 120 + 25, "%s\n%s\n", 
					sc->span.name, sc->span.desc);
	snprintf(card_configuration, 120 + 25, 
					"%sPort #%d -- FXO [%s]\n",
					card_configuration, 1, sc->span.alarms ? "off" : "on");
}



static int wcfxo_watchdog(struct zt_span *span, int event)
{
	printf("FXO: Restarting DMA\n");
	wcfxo_restart_dma(span->pvt);
	return 0;
}


static int wcfxo_open(struct zt_chan *chan)
{
	struct wcfxo_softc *sc = chan->pvt;
	if (sc->dead)
		return -ENODEV;
	sc->usecount++;
	return 0;
}

static int wcfxo_close(struct zt_chan *chan)
{
	struct wcfxo_softc *sc = chan->pvt;
	sc->usecount--;
	/* Actualy it is useless in BSD version but let it remains */
	/* If we're dead, release us now */
	if (!sc->usecount && sc->dead)
		wcfxo_release(sc);
	return 0;
}


static void
wcfxo_dma_map_addr(arg, segs, nseg, error)
		void *arg;
		bus_dma_segment_t *segs;
		int nseg, error;
{
		u_int32_t *paddr;
		
		paddr = arg;
		*paddr = segs->ds_addr;
}




static int wcfxo_hooksig(struct zt_chan *chan, zt_txsig_t txsig)
{
	struct wcfxo_softc *sc = chan->pvt;
	int reg=0;
	switch(txsig) {
	case ZT_TXSIG_START:
	case ZT_TXSIG_OFFHOOK:
		/* Take off hook and enable normal mode reception.  This must
		   be done in two steps because of a hardware bug. */
		reg = sc->readregs[0x5] & ~0x08;
		wcfxo_setreg(sc, 0x5, reg);

		reg = reg | 0x1;
		wcfxo_setreg(sc, 0x5, reg);
		sc->offhook = 1;
#ifdef  JAPAN
		sc->battery = 1;
		sc->battdebounce = BATT_DEBOUNCE;
		sc->ohdebounce = OH_DEBOUNCE;
#endif
		break;
	case ZT_TXSIG_ONHOOK:
		/* Put on hook and enable on hook line monitor */
		reg =  sc->readregs[0x5] & 0xfe;
		wcfxo_setreg(sc, 0x5, reg);

		reg = reg | 0x08;
		wcfxo_setreg(sc, 0x5, reg);

		sc->offhook = 0;
		/* Don't accept a ring for another 1000 ms */
		sc->ringdebounce = 1000;
#ifdef  JAPAN
		sc->ohdebounce = OH_DEBOUNCE;
#endif
		break;
	default:
		printf("wcfxo: Can't set tx state to %d\n", txsig);
	}
	if (debug)
	{
		printf("Setting hook state to %d (%02x)\n", txsig, reg);
	}
	return 0;
}


static int wcfxo_initialize(struct wcfxo_softc *sc)
{
	/* Zapata stuff */
	sprintf(sc->span.name, "WCFXO/%d", sc->pos);
	sprintf(sc->span.desc, "%s Board %d", sc->variety, sc->pos + 1);
	sprintf(sc->chan.name, "WCFXO/%d/%d", sc->pos, 0);
	sc->chan.sigcap = ZT_SIG_FXSKS | ZT_SIG_FXSLS | ZT_SIG_SF;
	sc->chan.chanpos = 1;
	sc->span.chans = &sc->chan;
	sc->span.channels = 1;
	sc->span.hooksig = wcfxo_hooksig;
	sc->span.open = wcfxo_open;
	sc->span.close = wcfxo_close;
	sc->span.flags = ZT_FLAG_RBS;
	sc->span.deflaw = ZT_LAW_MULAW;
	sc->span.watchdog = wcfxo_watchdog;
	/* init_waitqueue_head(&sc->span.maintq); */

	sc->span.pvt = sc;
	sc->chan.pvt = sc;
	if (zt_register(&sc->span, 0)) {
		printf("Unable to register span with zaptel\n");
		return -1;
	}
	return 0;
}


int 
wcfxo_hardware_init(struct wcfxo_softc * sc)
{
	/* Hardware stuff */
	unsigned char b;
	u_int32_t write_dma_addr;
	u_int32_t read_dma_addr;

	/* Reset PCI Interface chip and registers */
	WRITE1(WC_CNTL, 0x0e);
	if (sc->flags & FLAG_RESET_ON_AUX5) {
		/* Set hook state to on hook for when we switch.
		   Make sure reset is high */
		WRITE1(WC_AUXD, 0x34);
	} else {
		/* Set hook state to on hook for when we switch */
		WRITE1(WC_AUXD, 0x24);
	}
	/* Set all to outputs except AUX 4, which is an input */
	WRITE1(WC_AUXC, 0xef);

	/* Back to normal, with automatic DMA wrap around */
	WRITE1(WC_CNTL, 0x01);

	/* Make sure serial port and DMA are out of reset */
	/* WRITE1(READ1(WC_CNTL) & 0xf9, WC_CNTL); */
	b = READ1(WC_CNTL);
	b &= 0xf9;

	WRITE1(WC_CNTL, b);

	/* Configure serial port for MSB->LSB operation */
	if (sc->flags & FLAG_DOUBLE_CLOCK)
		WRITE1(WC_SERCTL, 0xc1);
	else
		WRITE1(WC_SERCTL, 0xc0);

	if (sc->flags & FLAG_USE_XTAL) {
		/* Use the crystal oscillator */
		WRITE1(WC_AUXFUNC, 0x04);
	}

	/* Delay FSC by 2 so it's properly aligned */
	WRITE1(WC_FSCDELAY, 0x2);

	/* Setup DMA Addresses */
	write_dma_addr = vtophys(sc->wc_ldata.wc_writechunk);
	
	WRITE4(WC_DMAWS, (unsigned int)write_dma_addr); /* Write start */
	WRITE4(WC_DMAWI, (unsigned int)write_dma_addr +
				ZT_CHUNKSIZE * 8 - 4	); /* Middle (interrupt) */
	WRITE4(WC_DMAWE, (unsigned int)write_dma_addr +
				ZT_CHUNKSIZE * 16 - 4); /* End */

	/* read_dma_addr = vtophys(sc->wc_ldata.wc_readchunk); */
	read_dma_addr =  vtophys(sc->wc_ldata.wc_readchunk);

	WRITE4(WC_DMARS, (unsigned int)read_dma_addr); /* Read start */
	WRITE4(WC_DMARI, (unsigned int)read_dma_addr +
				ZT_CHUNKSIZE * 8 - 4	); /* Middle (interrupt) */
	WRITE4(WC_DMARE, (unsigned int)read_dma_addr +
				ZT_CHUNKSIZE * 16 - 4); /* End */



	/* Clear interrupts */
	WRITE1(WC_INTSTAT, 0xff);
	return 0;
}

void 
wcfxo_enable_interrupts(struct wcfxo_softc * sc)
{
	/* Enable interrupts (we care about all of them) */
	WRITE1(WC_MASK0, 0x3f);
	/* No external interrupts */
	WRITE1(WC_MASK1, 0x00);
}



/*
 * Return identification string if this is device is ours.
 */
static int
wcfxo_probe(device_t dev)
{
	uint16_t		vid;
	uint16_t		devid;
	uint16_t		svid;
	struct wcfxo_ident 	*ident;

	vid	= pci_get_vendor(dev);
	devid 	= pci_get_device(dev);
	svid 	= pci_get_subvendor(dev);
	for (ident = wcfxo_ident_table; ident->device_info != NULL; ident++) {
		if ((ident->vid == vid) && (ident->did == devid) &&
			(ident->svid == svid || ident->svid == -1)) {

			printf("ZapTel device: vendor=%x device=%x subvendor=%x\n", 
				vid, devid, svid);
			device_set_desc(dev, ident->device_info->name);
			return (0);
		}
	}
	return (ENXIO);
}

static __inline void wcfxo_receiveprep(struct wcfxo_softc *sc, unsigned char ints)
{
	volatile int *readchunk;
	int x;
	int realreg;
	int realval;
	int sample;
	if (ints & 0x04)
		/* Read is at interrupt address.  Valid data is available at 
		 * normal offset 
		 */ 
		readchunk = sc->wc_ldata.wc_readchunk;
	else
		readchunk = sc->wc_ldata.wc_readchunk + ZT_CHUNKSIZE * 2;

	bus_dmamap_sync(sc->wc_ldata.wc_read_tag,
		sc->wc_ldata.wc_read_dmamap, BUS_DMASYNC_PREWRITE);

	/* Keep track of how quickly our peg alternates */
	sc->pegtimer+=ZT_CHUNKSIZE;
	sc->alt = 0;
	for (x=0;x<ZT_CHUNKSIZE;x++) {

		/* We always have a command to follow our signal.  */
		/* printf("reading... %d(flag=%d ignpreread=%d)\n",sc->oldregs[x].reg, 
			sc->oldregs[x].flags, sc->ignoreread); */
		if (sc->oldregs[x].flags == FLAG_READ && !sc->ignoreread) {
			realreg = wecareregs[(sc->regs[x].index + sc->regoffset) %
							(sizeof(wecareregs) / sizeof(wecareregs[0]))];
			realval = readchunk[(x << 1) +sc->alt];
			realval = (le32toh(realval) >> 16) & 0xff;

			if ((realval == 0x89) && (realreg != 0x9)) {
				/* Some sort of slippage, correct for it */
				while(realreg != 0x9) {
					/* Find register 9 */
					realreg = wecareregs[(sc->regs[x].index + ++sc->regoffset) %
										 (sizeof(wecareregs) / sizeof(wecareregs[0]))];
					sc->regoffset = sc->regoffset % (sizeof(wecareregs) / sizeof(wecareregs[0]));
				}
				if (debug)
					printf("New regoffset: %d\n", sc->regoffset);
			}
			/* Receive into the proper register */
			sc->readregs[realreg] = realval;
		}
		/* Look for pegging to indicate ringing */
		sample = (short)(le32toh(readchunk[(x << 1) + (1 - sc->alt)]) >> 16);
		if ((sample > 32000) && (sc->peg != 1)) {
			if ((sc->pegtimer < PEGTIME) && (sc->pegtimer > MINPEGTIME))
				sc->pegcount++;
			sc->pegtimer = 0;
			sc->peg = 1;
		} else if ((sample < -32000) && (sc->peg != -1)) {
			if ((sc->pegtimer < PEGTIME) && (sc->pegtimer > MINPEGTIME))
				sc->pegcount++;
			sc->pegtimer = 0;
			sc->peg = -1;
		}
		sc->chan.readchunk[x] = ZT_LIN2X((sample), (&sc->chan));
	}
	if (sc->pegtimer > PEGTIME) {
		/* Reset pegcount if our timer expires */
		sc->pegcount = 0;
	}
	/* Decrement debouncer if appropriate */
	if (sc->ringdebounce)
		sc->ringdebounce--;
	if (!sc->offhook && !sc->ringdebounce) {
		if (!sc->ring && (sc->pegcount > PEGCOUNT)) {
			/* It's ringing */
			if (debug)
				printf("RING!\n");
			zt_hooksig(&sc->chan, ZT_RXSIG_RING);
			sc->ring = 1;
		}
		if (sc->ring && !sc->pegcount) {
			/* No more ring */
			if (debug)
				printf("NO RING!\n");
			zt_hooksig(&sc->chan, ZT_RXSIG_OFFHOOK);
			sc->ring = 0;
		}
	}
	if (sc->ignoreread)
		sc->ignoreread--;

	/* Do the echo cancellation...  We are echo cancelling against
	   what we sent two chunks ago*/
	zt_ec_chunk(&sc->chan, sc->chan.readchunk, sc->lasttx);

	/* Receive the result */
	zt_receive(&sc->span);

	bus_dmamap_sync(sc->wc_ldata.wc_read_tag,
		sc->wc_ldata.wc_read_dmamap, BUS_DMASYNC_POSTWRITE);
}


static __inline void wcfxo_transmitprep(struct wcfxo_softc *sc, unsigned char ints)
{
	volatile int *writechunk;
	int x;
	int written=0;
	unsigned short cmd;

	/* if nothing to transmit, have to do the zt_transmit() anyway */
	if (!(ints & 3)) {
		/* Calculate Transmission */
		zt_transmit(&sc->span);
		return;
	}



	/* Remember what it was we just sent */
	memcpy(sc->lasttx, sc->chan.writechunk, ZT_CHUNKSIZE);

	if (ints & 0x01)  {
		/* Write is at interrupt address.  Start writing from normal offset */
		writechunk = sc->wc_ldata.wc_writechunk;
	} else {
		writechunk = sc->wc_ldata.wc_writechunk + ZT_CHUNKSIZE * 2;
	}

	zt_transmit(&sc->span);

	bus_dmamap_sync(sc->wc_ldata.wc_write_tag,
		sc->wc_ldata.wc_write_dmamap, BUS_DMASYNC_PREREAD);
	for (x=0;x<ZT_CHUNKSIZE;x++) {
		/* Send a sample, as a 32-bit word, and be sure to indicate that a command follows */
		if (sc->flags & FLAG_INVERTSER)
			writechunk[x << 1] = htole32(
				~((unsigned short)(ZT_XLAW(sc->chan.writechunk[x], (&sc->chan)))| 0x1) << 16
				);
		else
			writechunk[x << 1] = htole32(
				((unsigned short)(ZT_XLAW(sc->chan.writechunk[x], (&sc->chan)))| 0x1) << 16
				);

		/* We always have a command to follow our signal */
		if (!sc->regs[x].flags) {
			/* Fill in an empty register command with a read for a potentially useful register  */
			sc->regs[x].flags = FLAG_READ;
			sc->regs[x].reg = wecareregs[sc->readpos];
			sc->regs[x].index = sc->readpos;
			sc->readpos++;
			if (sc->readpos >= (sizeof(wecareregs) / sizeof(wecareregs[0]))) {
				sc->allread = 1;
				sc->readpos = 0;
			}
		}

		/* Prepare the command to follow it */
		switch(sc->regs[x].flags) {
		case FLAG_READ:
			cmd = (sc->regs[x].reg | 0x20) << 8;
			break;
		case FLAG_WRITE:
			cmd = (sc->regs[x].reg << 8) | (sc->regs[x].value & 0xff);
			written = 1;
			/* Wait at least four samples before reading */
			sc->ignoreread = 4;
			break;
		default:
			printf("wcfxo: Huh?  No read or write??\n");
			cmd = 0;
		}
		/* Setup the write chunk */
		if (sc->flags & FLAG_INVERTSER)
			writechunk[(x << 1) + 1] = htole32(~(cmd << 16));
		else
			writechunk[(x << 1) + 1] = htole32(cmd << 16);
	}

	/* printf("--end chunk\n"); */
	if (written)
		sc->readpos = 0;
	sc->wregcount = 0;


	for (x=0;x<ZT_CHUNKSIZE;x++) {
		/* Rotate through registers */
		sc->oldregs[x] = sc->regs[x];
		sc->regs[x].flags = FLAG_EMPTY;
	}


	bus_dmamap_sync(sc->wc_ldata.wc_write_tag,
		sc->wc_ldata.wc_write_dmamap, BUS_DMASYNC_POSTREAD);
}

static int wcfxo_init_daa(struct wcfxo_softc *sc)
{
	unsigned char reg15;
	/* This must not be called in an interrupt */
	/* We let things settle for a bit */
	tsleep(sc, PZERO | PCATCH, "waitintr", 10);

	/* Soft-reset it */
	wcfxo_setreg(sc, 0x1, 0x80);

	/* Let the reset go */
	/* set_current_state(TASK_UNINTERRUPTIBLE);
	schedule_timeout(1 + (ZT_CHUNKSIZE * HZ) / 800); */
	/* printf("+delay 2(%d)\n", (1 + (ZT_CHUNKSIZE * HZ) / 800)*10);
	DELAY((1 + (ZT_CHUNKSIZE * HZ) / 800)*10);
	printf("-delay 2\n"); */
	tsleep(sc, PZERO, "waitintr", (1 + (ZT_CHUNKSIZE*hz)/800)*10 );

	/* We have a clock at 18.432 Mhz, so N1=1, M1=2, CGM=0 */
	wcfxo_setreg(sc, 0x7, 0x0); /* This value is N1 - 1 */
	wcfxo_setreg(sc, 0x8, 0x1); /* This value is M1 - 1 */
	/* We want to sample at 8khz, so N2 = 9, M2 = 10 (N2-1, M2-1) */
	wcfxo_setreg(sc, 0x9, 0x89);

	/* Wait until the PLL's are locked. Time is between 100 uSec and 1 mSec */
	/*set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(1 + HZ/1000 + (ZT_CHUNKSIZE * HZ) / 800); */
	/* printf("delay 3\n");
	DELAY((1 + HZ/1000 + (ZT_CHUNKSIZE * HZ) / 800)*10);  */
	tsleep(sc, PZERO | PCATCH, "waitintr", 
		(1 + hz/1000 + (ZT_CHUNKSIZE * hz) / 800)*10);

	/* No additional ration is applied to the PLL and faster lock times
	 * are possible */
	wcfxo_setreg(sc, 0xa, 0x0);
	/* Enable off hook pin */
	wcfxo_setreg(sc, 0x5, 0x0a);
	if (monitor) {
		/* Enable ISOcap and external speaker and charge pump if present */
		wcfxo_setreg(sc, 0x6, 0x80);
	} else {
		/* Enable ISOcap and charge pump if present (leave speaker disabled) */
		wcfxo_setreg(sc, 0x6, 0xe0);
	}

	/* Wait a couple of jiffies for our writes to finish */
	/*set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(1 + (ZT_CHUNKSIZE * HZ) / 800);*/
	/* printf("delay 4\n");
	 DELAY((1 + (ZT_CHUNKSIZE * HZ) / 800)*10); */
	tsleep(sc, PZERO | PCATCH, "waitintr", (1 + (ZT_CHUNKSIZE * hz) / 800)*10);

	reg15 = 0;
	/* Go ahead and attenuate transmit signal by 6 db */
	if (quiet) {
		printf("wcfxo: Attenuating transmit signal for quiet operation\n");
		reg15 |= (quiet & 0x3) << 4;
	}

	if (boost) {
		printf("wcfxo: Boosting receive signal\n");
		reg15 |= (boost & 0x3);
	}

	wcfxo_setreg(sc, 0xf, 0x10);

	/* Didn't get it right.  Register 9 is still garbage */
	if (sc->readregs[0x9] != 0x89)
		return -1;
#if 0
	{ int x;
	int y;
	for (y=0;y<100;y++) {
		printf(" reg dump ====== %d ======\n", y);
		for (x=0;x<sizeof(wecareregs) / sizeof(wecareregs[0]);x++) {
			printf("daa: Reg %d: %02x\n", wecareregs[x], sc->readregs[wecareregs[x]]);
		}
		DELAY(100);
	} }
#endif  
	return 0;
}

static int wcfxo_set_daa_mode(struct wcfxo_softc *sc)
{
	/* Set country specific parameters (OHS, ACT, DCT, RZ, RT, LIM, VOL) */
	int reg16 = ((fxo_modes[opermode].ohs & 0x1) << 6) |
				((fxo_modes[opermode].act & 0x1) << 5) |
				((fxo_modes[opermode].dct & 0x3) << 2) |
				((fxo_modes[opermode].rz & 0x1) << 1) |
				((fxo_modes[opermode].rt & 0x1) << 0);
	int reg17 = ((fxo_modes[opermode].lim & 0x3) << 3);
	int reg18 = ((fxo_modes[opermode].vol & 0x3) << 3);

	wcfxo_setreg(sc, 16, reg16);
	wcfxo_setreg(sc, 17, reg17);
	wcfxo_setreg(sc, 18, reg18);


	/* Wait a couple of jiffies for our writes to finish */
	tsleep(sc, PZERO, "writefinish", 1 + (ZT_CHUNKSIZE * hz) / 800);

	printf("wcfxo: DAA mode is '%s'\n", fxo_modes[opermode].name);
	return 0;
}




static 
int wcfxo_dma_allocate(struct wcfxo_softc * sc)
{
	int error;

	/* write buffers */
#if __FreeBSD_version < 500000
	error = bus_dma_tag_create(NULL, 8, 0,
		BUS_SPACE_MAXADDR_32BIT, BUS_SPACE_MAXADDR, NULL, NULL,
		ZT_MAX_CHUNKSIZE * 2 * 2 * 2 * 4, 1, ZT_MAX_CHUNKSIZE * 2 * 2 * 2 * 4, 
			0, &sc->wc_ldata.wc_write_tag);
#else
	error = bus_dma_tag_create(NULL, 8, 0,
		BUS_SPACE_MAXADDR_32BIT, BUS_SPACE_MAXADDR, NULL, NULL,
		ZT_MAX_CHUNKSIZE * 2 * 2 * 2 * 4, 1, ZT_MAX_CHUNKSIZE * 2 * 2 * 2 * 4, 
			0, NULL, NULL, &sc->wc_ldata.wc_write_tag);
#endif

	if (error) {
		printf("wcfxo: failed to allocate writechunk dma tag\n");
		return 1;
	}

	error = bus_dmamem_alloc(sc->wc_ldata.wc_write_tag,
		(void **)&sc->wc_ldata.wc_writechunk, BUS_DMA_NOWAIT,
		&sc->wc_ldata.wc_write_dmamap);

	if (error) {
		printf("wcfxo: no memory for writedma buffers!\n");
		/* destroy tag */
		bus_dma_tag_destroy(sc->wc_ldata.wc_write_tag);
		sc->wc_ldata.wc_write_tag = NULL;
		return 1;
	}

	error = bus_dmamap_load(sc->wc_ldata.wc_write_tag,
		sc->wc_ldata.wc_write_dmamap, sc->wc_ldata.wc_writechunk,
		ZT_MAX_CHUNKSIZE * 2 * 2 * 2 * 4, wcfxo_dma_map_addr,
		&sc->wc_ldata.wc_write_dmaaddr, 0);

	if (error) {
			printf("wcfxo: can't load dma mapping");
			bus_dmamem_free(sc->wc_ldata.wc_write_tag, sc->wc_ldata.wc_writechunk, 
								sc->wc_ldata.wc_write_dmamap);
			/* destroy map */
			bus_dmamap_destroy(sc->wc_ldata.wc_write_tag, sc->wc_ldata.wc_write_dmamap);
			/* destroy tag */
			bus_dma_tag_destroy(sc->wc_ldata.wc_write_tag);
			sc->wc_ldata.wc_write_tag = NULL;
			return 1;
	}



	/* read buffers */
#if __FreeBSD_version < 500000
	error = bus_dma_tag_create(NULL, 8, 0,
		BUS_SPACE_MAXADDR_32BIT, BUS_SPACE_MAXADDR, NULL, NULL,
		ZT_MAX_CHUNKSIZE * 2 * 2 * 2 * 4, 1, ZT_MAX_CHUNKSIZE * 2 * 2 * 2 * 4, 
			0, &sc->wc_ldata.wc_read_tag);
#else
	error = bus_dma_tag_create(NULL, 8, 0,
		BUS_SPACE_MAXADDR_32BIT, BUS_SPACE_MAXADDR, NULL, NULL,
		ZT_MAX_CHUNKSIZE * 2 * 2 * 2 * 4, 1, ZT_MAX_CHUNKSIZE * 2 * 2 * 2 * 4, 
			0, NULL, NULL, &sc->wc_ldata.wc_read_tag);
#endif

	if (error) {
		printf("wcfxo: failed to allocate readchunk dma tag\n");
		return 1;
	}

	error = bus_dmamem_alloc(sc->wc_ldata.wc_read_tag,
		(void **)&sc->wc_ldata.wc_readchunk, BUS_DMA_NOWAIT,
		&sc->wc_ldata.wc_read_dmamap);

	if (error) {
		printf("wcfxo: no memory for readdma buffers!\n");
		/* destroy tag */
		bus_dma_tag_destroy(sc->wc_ldata.wc_read_tag);
		sc->wc_ldata.wc_read_tag = NULL;
		return 1;
	}

	error = bus_dmamap_load(sc->wc_ldata.wc_read_tag,
		sc->wc_ldata.wc_read_dmamap, sc->wc_ldata.wc_readchunk,
		ZT_MAX_CHUNKSIZE * 2 * 2 * 2 * 4, wcfxo_dma_map_addr,
		&sc->wc_ldata.wc_read_dmaaddr, BUS_DMA_NOWAIT);

	if (error) {
			printf("wcfxo: can't load dma mapping");
			bus_dmamem_free(sc->wc_ldata.wc_read_tag, sc->wc_ldata.wc_readchunk, 
								sc->wc_ldata.wc_read_dmamap);
			/* destroy map */
			bus_dmamap_destroy(sc->wc_ldata.wc_read_tag, sc->wc_ldata.wc_read_dmamap);
			/* destroy tag */
			bus_dma_tag_destroy(sc->wc_ldata.wc_read_tag);
			sc->wc_ldata.wc_read_tag = NULL;
			return 1;
	}


	return 0;
}


static void wcfxo_interrupt(void * arg)
{
	struct wcfxo_softc *sc = arg;
	unsigned char ints = 0xff;
	unsigned char b;
#ifdef DEBUG_RING
	static int oldb = 0;
	static int oldcnt = 0;
#endif

	ints = READ1(WC_INTSTAT);
	WRITE1(WC_INTSTAT, ints);

	if (!ints)
		return;


	if (ints & 0x0c) {  /* if there is a rx interrupt pending */
		/* transmitprep looks to see if there is anything to transmit
		   and returns by itself if there is nothing */
		wcfxo_receiveprep(sc, ints);
		wcfxo_transmitprep(sc, ints);
		/* wakeup(&qqq); */
	}

	if (ints & 0x10) {
		printf("FXO PCI Master abort\n");
		/* Stop DMA andlet the watchdog start it again */
		wcfxo_stop_dma(sc);
		return;
	}

	if (ints & 0x20) {
		printf("PCI Target abort\n");
		return;
	}

	if (batterycheck) {
		/* Check for BATTERY from register and debounce for 8 ms */
		b = sc->readregs[0xc] & 0xf;
		if (!b) {
			sc->nobatttimer++;
#if 0
			if (sc->battery)
				printf("Battery loss: %d (%d debounce)\n", b, sc->battdebounce);
#endif
			if (sc->battery && !sc->battdebounce) {
				if (debug)
					printf("NO BATTERY!\n");
				sc->battery =  0;
#ifdef  JAPAN
				if ((!sc->ohdebounce) && sc->offhook) {
					zt_hooksig(&sc->chan, ZT_RXSIG_ONHOOK);
					if (debug)
						printf("Signalled On Hook\n");
#ifdef  ZERO_BATT_RING
					sc->onhook++;
#endif
				}
#else
				zt_hooksig(&sc->chan, ZT_RXSIG_ONHOOK);
#endif
				sc->battdebounce = BATT_DEBOUNCE;
			} else if (!sc->battery)
				sc->battdebounce = BATT_DEBOUNCE;
			if ((sc->nobatttimer > 5000) &&
#ifdef  ZERO_BATT_RING
				!(sc->readregs[0x05] & 0x04) &&
#endif
				(!sc->span.alarms)) {
				sc->span.alarms = ZT_ALARM_RED;
				zt_alarm_notify(&sc->span);
				wcfxo_generate_description(sc);
			}
		} else if (b == 0xf) {
			if (!sc->battery && !sc->battdebounce) {
				if (debug)
					printf("BATTERY!\n");
#ifdef  ZERO_BATT_RING
				if (sc->onhook) {
					sc->onhook = 0;
					zt_hooksig(&sc->chan, ZT_RXSIG_OFFHOOK);
					if (debug)
						printf("Signalled Off Hook\n");
				}
#else
				zt_hooksig(&sc->chan, ZT_RXSIG_OFFHOOK);
#endif
				sc->battery = 1;
				sc->nobatttimer = 0;
				sc->battdebounce = BATT_DEBOUNCE;
				if (sc->span.alarms) {
					sc->span.alarms = 0;
					zt_alarm_notify(&sc->span);
					wcfxo_generate_description(sc);
				}
			} else if (sc->battery)
				sc->battdebounce = BATT_DEBOUNCE;
		} else {
			/* It's something else... */
				sc->battdebounce = BATT_DEBOUNCE;
		}

		if (sc->battdebounce)
			sc->battdebounce--;
#ifdef  JAPAN
		if (sc->ohdebounce)
			sc->ohdebounce--;
#endif

	} else
	{
		/* We don't want to check battery status
		 * It could be usefull for old pulse telephone lines 
		 * (like here, in Kiev)to prevent driver to make wrong hangup
		 * So we should just reset all alarms on a line
		 */
		if(sc->span.alarms) {
			sc->span.alarms = 0;
			zt_alarm_notify(&sc->span);
			wcfxo_generate_description(sc);
		}
	}

}

static void wcfxo_sysctl_init()
{
	sysctl_ctx_init(&wcfxo_clist);

	wcfxo_root = SYSCTL_ADD_NODE(&wcfxo_clist, SYSCTL_CHILDREN(zt_root),
		OID_AUTO, "wcfxo", CTLFLAG_RD, 0, "wcfxo driver settings");

	SYSCTL_ADD_INT(&wcfxo_clist, SYSCTL_CHILDREN(wcfxo_root),
		OID_AUTO, "debug", CTLFLAG_RW, &debug, 0, 
		"make X100* driver more verbose");

	SYSCTL_ADD_INT(&wcfxo_clist, SYSCTL_CHILDREN(wcfxo_root),
		OID_AUTO, "monitor", CTLFLAG_RW, &monitor, 0, 
		"monitor");

	/* TODO: make this parameters available not only load tim
	 * but run-time too 
	 
	SYSCTL_ADD_INT(&wcfxo_clist, SYSCTL_CHILDREN(wcfxo_root),
		OID_AUTO, "quiet", CTLFLAG_RW, &quiet, 0, 
		"Attenuate transmit signal by 6 db");

	SYSCTL_ADD_INT(&wcfxo_clist, SYSCTL_CHILDREN(wcfxo_root),
		OID_AUTO, "boost", CTLFLAG_RW, &boost, 0, 
		"Boosting receive signal");
		

	SYSCTL_ADD_INT(&wcfxo_clist, SYSCTL_CHILDREN(wcfxo_root),
		OID_AUTO, "opermode", CTLFLAG_RW, &opermode, 0, 
		"Country specific operation mode");
	*/

	SYSCTL_ADD_INT(&wcfxo_clist, SYSCTL_CHILDREN(wcfxo_root),
		OID_AUTO, "batterycheck", CTLFLAG_RW, &batterycheck, 0, 
		"Check line state (hangup detection)");

}

void wcfxo_sysctl_destroy()
{
	sysctl_ctx_free(&wcfxo_clist);
}



/* Attach function is only called if the probe is successful */
static int
wcfxo_attach(device_t dev)
{
	u_int16_t 				error;
	u_int32_t 				command;
	struct wcfxo_softc 		*sc;
#if __FreeBSD_version < 500000
	dev_t					d;
#else
	struct cdev				* d;
#endif
	int 					unit, rid;
	struct wcfxo_ident 		*ident;
	uint16_t				vid;
	uint16_t				devid;
	uint16_t				svid;

	vid	= pci_get_vendor(dev);
	devid 	= pci_get_device(dev);
	svid 	= pci_get_subvendor(dev);
	for (ident = wcfxo_ident_table; ident->device_info != NULL; ident++)
		if ((ident->vid == vid) && (ident->did == devid) &&
			(ident->svid == svid || ident->svid == -1))
				break;

	if(ident->device_info == 0)
	{
		printf("Card info not found\n");
		error = ENXIO;
		goto fail;
	}

	sc = (struct wcfxo_softc *)device_get_softc(dev);
	unit = device_get_unit(dev);

	printf("ZapTel Attach for wcfxo%d: deviceID : 0x%x\n",
		unit, pci_get_vendor(dev));

	d = make_dev(&wcfxo_cdevsw, unit, UID_ROOT,
		GID_WHEEL, 0644, "wcfxo%d", unit);

	sc->d = d;

	/*
	 * Enable bus mastering.
	 */
	pci_enable_busmaster(dev);
	pci_enable_io(dev, SYS_RES_IOPORT);
	pci_enable_io(dev, SYS_RES_MEMORY);
	command = pci_read_config(dev, PCIR_COMMAND, 4);

	if (!(command & PCIM_CMD_PORTEN) && !(command & PCIM_CMD_MEMEN)) {
		printf("wcfxo%d: failed to enable I/O ports and memory mappings!\n", unit);
		error = ENXIO;
		goto fail;
	}



	sc->wc_ioport_rid = PCIR_BAR(0);
	sc->wc_base_addr = bus_alloc_resource(dev, SYS_RES_IOPORT, 
		&sc->wc_ioport_rid, 0ul, ~0ul, 1, RF_ACTIVE);

	if (!sc->wc_base_addr) {
			error = ENOMEM;
			goto fail;
	}
	sc->wc_base_addr_tag	= rman_get_bustag(sc->wc_base_addr);
	sc->wc_base_addr_handle = rman_get_bushandle(sc->wc_base_addr);


	sc->pos = unit;

	sc->variety = ident->device_info->name;
	/* sc->flags = FLAG_INVERTSER | FLAG_USE_XTAL | FLAG_DOUBLE_CLOCK ; */
	sc->flags = ident->device_info->flags;

	/* Alocate memory for two zt chunks - recieve and transmit */
	if(wcfxo_dma_allocate(sc))
	{
		printf("wcfxo: Unable to intialize DMA memory\n");
		wcfxo_detach(dev);
		return ENOMEM;
	}

	if (wcfxo_initialize(sc)) {
		printf("wcfxo: Unable to intialize modem\n");
		if (sc->wc_base_addr != NULL)
			if(bus_release_resource(dev, SYS_RES_IOPORT,
					sc->wc_ioport_rid, sc->wc_base_addr) != 0)
				printf("can't release resource(no parent)");
		return EIO;
	}

	/* Request IRQ resource */
	rid = 0x0;
	sc->wc_irq = bus_alloc_resource(dev, SYS_RES_IRQ, &rid,
		0, ~0, 1, RF_SHAREABLE | RF_ACTIVE);
	if (sc->wc_irq == NULL) {
		printf("IRQ allocation failed!\n");
		error = ENXIO;
		wcfxo_detach(dev);
		goto fail;
	}

	/* Now we should set up the interrupt handler */

	error = bus_setup_intr(dev, sc->wc_irq, INTR_TYPE_MISC,
		(void (*)(void *))wcfxo_interrupt, sc, &(sc->irq_handler));
	if (error) {
		printf("Couldn't set up irq\n");
		wcfxo_detach(dev);
		goto fail;
	}

	sc->wc_irq_bt = rman_get_bustag(sc->wc_irq);
	sc->wc_irq_bh = rman_get_bushandle(sc->wc_irq);



	wcfxo_hardware_init(sc);
	/* Enable interrupts */
	wcfxo_enable_interrupts(sc);
	/* Start DMA */
	wcfxo_start_dma(sc);

	/* Initialize Write/Buffers to all blank data */
	memset((void *)(sc->wc_ldata.wc_writechunk), 0, 
			ZT_MAX_CHUNKSIZE * 2 * 2 * 2 * 4);

	memset((void *)(sc->wc_ldata.wc_readchunk), 0, 
			ZT_MAX_CHUNKSIZE * 2 * 2 * 2 * 4);

	/* Initialize DAA (after it's started) */
	if (wcfxo_init_daa(sc)) {
		printf("Failed to initailize DAA, giving up...\n");
		wcfxo_detach(dev);
		error = EIO;
		goto fail;
	}

	wcfxo_set_daa_mode(sc);
	wcfxo_generate_description(sc);

	printf("Found a Wildcard FXO: %s\n", sc->variety);
	printf("ZapTel device loaded.\n");

	return 0;

fail:
	printf("We failed: %d\n", error);
	return error;
}

/* Detach device. */
static int
wcfxo_detach(device_t dev)
{
	struct wcfxo_softc * sc;

	sc = (struct wcfxo_softc *)device_get_softc(dev);
	if(sc)
	{
		/* we're busy - there is open istances exists */
		if(sc->usecount)
			return EBUSY;
		/* In case hardware is still there */
		wcfxo_disable_interrupts(sc);
		/* way to settle down things */
		DELAY(100);

		/* Stop any DMA */
		wcfxo_stop_dma(sc);
		wcfxo_reset_tdm(sc);


		/* Immediately free resources */
		bus_generic_detach(dev);

		if (sc->wc_irq)
		{
			bus_teardown_intr(dev, sc->wc_irq, sc->irq_handler);
			bus_release_resource(dev, SYS_RES_IRQ, 0, sc->wc_irq);
		}

		if (sc->wc_base_addr != NULL)
			if(bus_release_resource(dev, SYS_RES_IOPORT,
				sc->wc_ioport_rid, sc->wc_base_addr)!= 0)
					printf("Can't release base addr");


		if (sc->wc_ldata.wc_read_tag) {
			bus_dmamap_unload(sc->wc_ldata.wc_read_tag,
				sc->wc_ldata.wc_read_dmamap);
			bus_dmamem_free(sc->wc_ldata.wc_read_tag, sc->wc_ldata.wc_readchunk,
			sc->wc_ldata.wc_read_dmamap);
			bus_dma_tag_destroy(sc->wc_ldata.wc_read_tag);
		}
	
		if (sc->wc_ldata.wc_write_tag) {
			bus_dmamap_unload(sc->wc_ldata.wc_write_tag,
				sc->wc_ldata.wc_write_dmamap);
			bus_dmamem_free(sc->wc_ldata.wc_write_tag, sc->wc_ldata.wc_writechunk,
				sc->wc_ldata.wc_write_dmamap);
			bus_dma_tag_destroy(sc->wc_ldata.wc_write_tag);
		}

		/* Reset PCI chip and registers */
		WRITE1(WC_CNTL, 0x0e);

		/* Release span, possibly delayed */
		if (!sc->usecount)
			wcfxo_release(sc);
		else
			sc->dead = 1;

		if(sc->d)
			destroy_dev(sc->d);

		if(card_configuration)
		{
			free(card_configuration, M_DEVBUF);
			card_configuration = 0;
		}
		printf("ZapTel detach!\n");
	}
	return (0);
}

/* Called during system shutdown after sync. */

static int
wcfxo_shutdown(device_t dev)
{

	printf("ZapTel shutdown!\n");
	return (0);
}

/*
 * Device suspend routine.
 */
static int
wcfxo_suspend(device_t dev)
{

	printf("ZapTel suspend!\n");
	return (0);
}

/*
 * Device resume routine.
 */

static int
wcfxo_resume(device_t dev)
{

	printf("ZapTel resume!\n");
	return (0);
}

/* 
 * syscall routines 
 * We will use them to emulate Linux /proc behaviour --
 * they will give detailed info about card installed
 */ 
#if __FreeBSD_version < 500000
int wcfxo_sys_open(dev_t dev, int flag, int devtype, struct proc *threadp)
#else
int wcfxo_sys_open(struct cdev * dev, int flag, int mode, struct thread *threadp)
#endif
{
	return 0;
}

#if __FreeBSD_version < 500000
int wcfxo_sys_close(dev_t dev, int flags, int fmt, struct proc *threadp)
#else
int wcfxo_sys_close(struct cdev * dev, int flags, int fmt, struct thread *threadp)
#endif
{
	return 0;
}

#if __FreeBSD_version < 500000
int
wcfxo_sys_read(dev_t dev, struct uio *uio, int flag)
#else
int
wcfxo_sys_read(struct cdev *dev, struct uio *uio, int flag)
#endif
{
	int size;
	/* end of file */
	if(uio->uio_offset < 0)
		return -1;

	if(!card_configuration)
		return 0;
	
    if(uio->uio_offset >= strlen(card_configuration))
		return 0;

	size = MIN(uio->uio_resid, strlen(card_configuration) - uio->uio_offset);
	return uiomove(card_configuration +  uio->uio_offset, size, uio);
}



static int
wcfxo_load(module_t mod, int cmd, void *arg)
{
    int err = 0;
	switch (cmd) {
		case MOD_LOAD:
    		err = driver_module_handler(mod, cmd, arg);
			if(!err)
				wcfxo_sysctl_init();
		break;          /* Success*/
     	
		case MOD_UNLOAD:
    		err = driver_module_handler(mod, cmd, arg);
			if(!err)
				wcfxo_sysctl_destroy();
		break;          /* Success */
 
   		default: 
			err = EINVAL;
 		break;
	}
	return err;
}

static device_method_t wcfxo_methods[] = {
	/* Device interface */
	DEVMETHOD(device_probe,	 	wcfxo_probe),
	DEVMETHOD(device_attach,	wcfxo_attach),
	DEVMETHOD(device_detach,	wcfxo_detach),
	DEVMETHOD(device_shutdown,  wcfxo_shutdown),
	DEVMETHOD(device_suspend,   wcfxo_suspend),
	DEVMETHOD(device_resume,	wcfxo_resume),

	{ 0, 0 }
};

static driver_t wcfxo_driver = {
	"wcfxo",
	wcfxo_methods,
	sizeof(struct wcfxo_softc)
};

static devclass_t wcfxo_devclass;

#if __FreeBSD_version < 500000
static driver_t *wcfxo_pci_driver_list = { &wcfxo_driver };
static struct driver_module_data wcfxo_pci_driver_mod = { 
	0, 0,
	"pci",
	&wcfxo_pci_driver_list,
	sizeof(wcfxo_pci_driver_list)/sizeof(wcfxo_pci_driver_list[0]),
	&wcfxo_devclass
};

static moduledata_t wcfxo_pci_mod = {
        "pci/wcfxo",
        wcfxo_load,
        &wcfxo_pci_driver_mod
};

#else


static struct driver_module_data wcfxo_pci_driver_mod = { 
	0, 0,
	"pci",
	(kobj_class_t) &wcfxo_driver,
	&wcfxo_devclass
};

static moduledata_t wcfxo_pci_mod = {
        "pci/wcfxo",
        wcfxo_load,
        &wcfxo_pci_driver_mod
};

#endif
MODULE_DEPEND(wcfxo, zaptel, 1, 1, 1);
/* DRIVER_MODULE(wcfxo, pci, wcfxo_driver, wcfxo_devclass, 0, 0); */
DECLARE_MODULE(wcfxo, wcfxo_pci_mod, SI_SUB_DRIVERS, SI_ORDER_MIDDLE);
