/*
 * $Id: wcfxs.c 101 2005-03-07 21:36:56Z riz $
 *
 *
 * NetBSD driver for Wildcard TDM400P FXS/FXO board from Digium, Inc.
 *
 * Original code written by:
 *	Mark Spencer <markster@linux-support.net>
 *	Matthew Fredrickson <creslin@linux-support.net>
 *
 * Copyright (C) 2001, Linux Support Services, Inc.
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * FreeBSD port by
 *	Alexander Timoshenko <gonzo@portaone.com>, PortaOne software
 *
 * NetBSD port, and code cleanup and rewriting 
 * 	Copyright (c) 2004, Red Crow Group LLC
 *	written by Jeff Rizzo <riz@redcrowgroup.com>
 */

#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$Id: wcfxs.c 101 2005-03-07 21:36:56Z riz $");

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/proc.h>
#include <sys/uio.h>
#include <sys/malloc.h>
#include <sys/device.h>
#include <sys/conf.h>

#include <uvm/uvm_extern.h>	/* for PAGE_SIZE */

#include <dev/pci/pcireg.h>
#include <dev/pci/pcivar.h>
#include <dev/pci/pcidevs.h>

#ifdef _LKM
#include <zaptel.h>
#include "wcfxsreg.h"
#include "wcfxsvar.h"
#include "proslic.h"
#else
#include <dev/zaptel/zaptel/zaptel.h>
#include <dev/zaptel/wcfxs-netbsd/wcfxsreg.h>
#include <dev/zaptel/wcfxs-netbsd/wcfxsvar.h>
#include <dev/zaptel/wcfxs-netbsd/proslic.h>
#endif

#define FXS_LOCK(lock) {};
#define FXS_UNLOCK(lock) {};
#define AUDIO_RINGCHECK 1 
#undef AUDIO_RINGCHECK
MALLOC_DEFINE(M_TDM, "wcfxs", "wcfxs interface data structures");

int wcfxs_match(struct device *, struct cfdata *, void *);
static void wcfxs_attach(struct device *, struct device *, void *);
static int wcfxs_detach(struct device *, int);

struct device *zaptel_attach_mi(struct device *);
int wcfxs_open(struct zt_chan *chan);
static int wcfxs_close(struct zt_chan *chan);
static int wcfxs_ioctl(struct zt_chan *chan, unsigned int cmd, caddr_t data);
static int wcfxs_watchdog(struct zt_span *span, int event);

static int wcfxs_intr(void *);
static int wcfxs_initialize(struct wcfxs_softc *);
static int wcfxs_init_proslic(struct wcfxs_softc *, int, int, int, int);
static int wcfxs_init_voicedaa(struct wcfxs_softc *, int, int, int, int);
static void wcfxs_voicedaa_check_hook(struct wcfxs_softc *sc, int card);
static void wcfxs_proslic_check_hook(struct wcfxs_softc *sc, int card);
static void wcfxs_proslic_recheck_sanity(struct wcfxs_softc *sc, int card);

static void wcfxs_stop_dma(struct wcfxs_softc *sc);
static void wcfxs_reset_tdm(struct wcfxs_softc *sc);
static void wcfxs_restart_dma(struct wcfxs_softc *sc);
static void wcfxs_release(struct wcfxs_softc *sc);
static void __wcfxs_setcreg(struct wcfxs_softc *, unsigned char,
    unsigned char);
static unsigned char __wcfxs_getcreg(struct wcfxs_softc *, unsigned char);
void wcfxs_generate_description(struct wcfxs_softc * sc);
static int wcfxs_hooksig(struct zt_chan *chan, zt_txsig_t txsig);

int wcfxsdebug = 1;
#ifdef DEBUG
#define DPRINTF(m, x)   if (wcfxsdebug >= (m)) printf x
#define DEBUG_DO(m, x)	if (wcfxsdebug >= (m)) do { x ; } while (0)
#else
#define DPRINTF(m, x)   /* nop */
#define DEBUG_DO(m, x)	/* nop */
#endif
static int robust = 0;
static char * card_configuration = NULL;

static alpha  indirect_regs[] =
{
{0,"DTMF_ROW_0_PEAK",0x55C2},
{1,"DTMF_ROW_1_PEAK",0x51E6},
{2,"DTMF_ROW2_PEAK",0x4B85},
{3,"DTMF_ROW3_PEAK",0x4937},
{4,"DTMF_COL1_PEAK",0x3333},
{5,"DTMF_FWD_TWIST",0x0202},
{6,"DTMF_RVS_TWIST",0x0202},
{7,"DTMF_ROW_RATIO_TRES",0x0198},
{8,"DTMF_COL_RATIO_TRES",0x0198},
{9,"DTMF_ROW_2ND_ARM",0x0611},
{10,"DTMF_COL_2ND_ARM",0x0202},
{11,"DTMF_PWR_MIN_TRES",0x00E5},
{12,"DTMF_OT_LIM_TRES",0x0A1C},
{13,"OSC1_COEF",0x7B30},
{14,"OSC1X",0x0063},
{15,"OSC1Y",0x0000},
{16,"OSC2_COEF",0x7870},
{17,"OSC2X",0x007D},
{18,"OSC2Y",0x0000},
{19,"RING_V_OFF",0x0000},
{20,"RING_OSC",0x7EF0},
{21,"RING_X",0x0160},
{22,"RING_Y",0x0000},
{23,"PULSE_ENVEL",0x2000},
{24,"PULSE_X",0x2000},
{25,"PULSE_Y",0x0000},
/* {26,"RECV_DIGITAL_GAIN",0x4000},	// playback volume set lower */
{26,"RECV_DIGITAL_GAIN",0x2000},	/* playback volume set lower */
{27,"XMIT_DIGITAL_GAIN",0x4000},
/*/{27,"XMIT_DIGITAL_GAIN",0x2000}, */
{28,"LOOP_CLOSE_TRES",0x1000},
{29,"RING_TRIP_TRES",0x3600},
{30,"COMMON_MIN_TRES",0x1000},
{31,"COMMON_MAX_TRES",0x0200},
{32,"PWR_ALARM_Q1Q2",0x07C0},
{33,"PWR_ALARM_Q3Q4",0x2600},
{34,"PWR_ALARM_Q5Q6",0x1B80},
{35,"LOOP_CLOSURE_FILTER",0x8000},
{36,"RING_TRIP_FILTER",0x0320},
{37,"TERM_LP_POLE_Q1Q2",0x008C},
{38,"TERM_LP_POLE_Q3Q4",0x0100},
{39,"TERM_LP_POLE_Q5Q6",0x0010},
{40,"CM_BIAS_RINGING",0x0C00},
{41,"DCDC_MIN_V",0x0C00},
{42,"DCDC_XTRA",0x1000},
{43,"LOOP_CLOSE_TRES_LOW",0x1000},
};

CFATTACH_DECL(wcfxs, sizeof(struct wcfxs_softc), 
	wcfxs_match, wcfxs_attach, wcfxs_detach, NULL);

/* descriptions for the boards which might match */
const static struct wcfxs_desc wcfxs = { "Wildcard S400P Prototype", 0 };
const static struct wcfxs_desc wcfxse = { "Wildcard TDM400P REV E/F", 0 };
const static struct wcfxs_desc wcfxsh = { "Wildcard TDM400P REV H", 0 };

/* board identification table */
static struct wcfxs_ident wcfxs_ident_table[] = {
	/* vendor, device, subvendor, card info struct */
	{ 0xe159, 0x0001, 0xa159, &wcfxs },
	{ 0xe159, 0x0001, 0xe159, &wcfxs },
	{ 0xe159, 0x0001, 0xb100, &wcfxse },
	{ 0xe159, 0x0001, 0xa9fd, &wcfxsh },
	{ 0xe159, 0x0001, 0xa8fd, &wcfxsh },
	{ 0xe159, 0x0001, 0xa81d, &wcfxsh },
	{ 0xe159, 0x0001, 0xa800, &wcfxsh },
	{ 0xe159, 0x0001, 0xa801, &wcfxsh },
	{ 0xe159, 0x0001, 0xa901, &wcfxsh },
	{ 0xe159, 0x0001, 0xa908, &wcfxsh },
	{ 0xe159, 0x0001, -1, &wcfxsh },
	{ 0, 0, 0, NULL }
};

#ifndef __FreeBSD__
/*
 * Compute number of ticks in the specified amount of time.
 * (code from FreeBSD)
 */
int tvtohz(struct timeval *);
int
tvtohz(tv)
	struct timeval *tv;
{
	register unsigned long ticks;
	register long sec, usec;

	/*
	 * If the number of usecs in the whole seconds part of the time
	 * difference fits in a long, then the total number of usecs will
	 * fit in an unsigned long.  Compute the total and convert it to
	 * ticks, rounding up and adding 1 to allow for the current tick
	 * to expire.  Rounding also depends on unsigned long arithmetic
	 * to avoid overflow.
	 *
	 * Otherwise, if the number of ticks in the whole seconds part of
	 * the time difference fits in a long, then convert the parts to
	 * ticks separately and add, using similar rounding methods and
	 * overflow avoidance.  This method would work in the previous
	 * case but it is slightly slower and assumes that hz is integral.
	 *
	 * Otherwise, round the time difference down to the maximum
	 * representable value.
	 *
	 * If ints have 32 bits, then the maximum value for any timeout in
	 * 10ms ticks is 248 days.
	 */
	sec = tv->tv_sec;
	usec = tv->tv_usec;
	if (usec < 0) {
		sec--;
		usec += 1000000;
	}
	if (sec < 0) {
#ifdef DIAGNOSTIC
		if (usec > 0) {
			sec++;
			usec -= 1000000;
		}
		printf("tvotohz: negative time difference %ld sec %ld usec\n",
		       sec, usec);
#endif
		ticks = 1;
	} else if (sec <= LONG_MAX / 1000000)
		ticks = (sec * 1000000 + (unsigned long)usec + (tick - 1))
			/ tick + 1;
	else if (sec <= LONG_MAX / hz)
		ticks = sec * hz
			+ ((unsigned long)usec + (tick - 1)) / tick + 1;
	else
		ticks = LONG_MAX;
	if (ticks > INT_MAX)
		ticks = INT_MAX;
	return ((int)ticks);
}

#endif

int
wcfxs_match(struct device *parent, struct cfdata *match, void *aux)
{
	struct pci_attach_args *pa = aux;
	pcireg_t subdev;
	struct wcfxs_ident *id;

	/* read the subdevice information */
	subdev = pci_conf_read(pa->pa_pc, pa->pa_tag, PCI_SUBSYS_ID_REG);

	for (id = wcfxs_ident_table; id->device_info != NULL; id++) {
		if (id->vendor == PCI_VENDOR(pa->pa_id) &&
			id->product == PCI_PRODUCT(pa->pa_id) &&
			(id->subvendor == PCI_VENDOR(subdev) ||
			id->subvendor == (pci_vendor_id_t) -1)) {

			printf("FXS device: ");
			printf("vendor=%x ",PCI_VENDOR(pa->pa_id));
			printf("product=%x ",PCI_PRODUCT(pa->pa_id));
			printf("subvendor=%x\n",PCI_VENDOR(subdev));
			return (1);
		}
	}
	return (0);
}

/* wcfxs_dma_malloc() returns 0 on failure, 1 on success */
static int
wcfxs_dma_malloc(struct wcfxs_softc *sc, struct wcfxs_dma *dma)
{
	int error;

	/* set up read buffer first */
	if ((error = bus_dmamem_alloc(sc->sc_dmat, WCFXS_SIZE, PAGE_SIZE, 0,
	    &dma->dma_seg, 1, &dma->dma_nseg, BUS_DMA_NOWAIT)) != 0) {
		aprint_error(
		    "%s: unable to allocate readchunk, error = %d\n",
		    sc->sc_dv.dv_xname, error);
		goto fail_0;
	}

	if ((error = bus_dmamem_map(sc->sc_dmat, &dma->dma_seg,
	    dma->dma_nseg, WCFXS_SIZE, (caddr_t *)&dma->chunk,
	    BUS_DMA_NOWAIT)) != 0) {
		aprint_error(
		    "%s: unable to map read chunk, error = %d\n",
		    sc->sc_dv.dv_xname, error);
		goto fail_1;
	}

	if ((error = bus_dmamap_create(sc->sc_dmat, WCFXS_SIZE, 1, WCFXS_SIZE,
								   0, 0, &dma->dma_map)) != 0) {
		aprint_error("%s: unable to create read DMA map, "
		    "error = %d\n",sc->sc_dv.dv_xname, error);
		goto fail_2;
	}

	if ((error = bus_dmamap_load(sc->sc_dmat, dma->dma_map,
	     dma->chunk, WCFXS_SIZE, NULL, BUS_DMA_NOWAIT)) != 0) {
		aprint_error(
		    "%s: unable to load read DMA map, error = %d\n",
		    sc->sc_dv.dv_xname, error);
		goto fail_3;
	}

	/* success! */
	return 1;

 fail_3:
	bus_dmamap_destroy(sc->sc_dmat, dma->dma_map);
 fail_2:
	bus_dmamem_unmap(sc->sc_dmat, (caddr_t)dma->chunk, WCFXS_SIZE);
 fail_1:
	bus_dmamem_free(sc->sc_dmat, &dma->dma_seg,
	    dma->dma_nseg);
 fail_0:
	return 0;
}

static void
wcfxs_dma_free(struct wcfxs_softc *sc, struct wcfxs_dma *dma)
{
	int size = ZT_MAX_CHUNKSIZE * 2 * 2 * 2 * 4; /* XXX magic */

	bus_dmamap_destroy(sc->sc_dmat, dma->dma_map);
	bus_dmamem_unmap(sc->sc_dmat, (caddr_t)dma->chunk, size);
	bus_dmamem_free(sc->sc_dmat, &dma->dma_seg,
	    dma->dma_nseg);
}

/*
 * wcfxs_initialize() - set up various bits of the softc
 */
static int
wcfxs_initialize(struct wcfxs_softc *sc)
{
	int x;

	/* set up the Zaptel bits */
	sprintf(sc->sc_span.name, "WCTDM/%d", sc->sc_pos);
	sprintf(sc->sc_span.desc, "%s Board %d", sc->sc_variety,
	    sc->sc_pos + 1);
	sc->sc_span.deflaw = ZT_LAW_MULAW;
	for (x = 0; x < sc->sc_cards; x++) {
	    sprintf(sc->sc_chans[x].name, "WCTDM/%d/%d",sc->sc_pos, x);
	    sc->sc_chans[x].sigcap = ZT_SIG_FXOKS | ZT_SIG_FXOLS |
		ZT_SIG_FXOGS | ZT_SIG_SF | ZT_SIG_EM | ZT_SIG_FXSKS |
		ZT_SIG_FXSLS;
	    sc->sc_chans[x].chanpos = x + 1;
	    sc->sc_chans[x].pvt = sc;
	}
	sc->sc_span.chans = sc->sc_chans;
	sc->sc_span.channels = sc->sc_cards;
	sc->sc_span.hooksig = wcfxs_hooksig;
	sc->sc_span.open = wcfxs_open;
	sc->sc_span.close = wcfxs_close;
	sc->sc_span.flags = ZT_FLAG_RBS;
	sc->sc_span.ioctl = wcfxs_ioctl;
	sc->sc_span.watchdog = wcfxs_watchdog;
	/* TODO: waitq init_waitqueue_head(&sc->sc_span.maintq); */

	sc->sc_span.pvt = sc;
	if (zt_register(&sc->sc_span, 0)) {
	    printf("Unable to register span with zaptel\n");
	    return -1;
	}
	return 0;
}

static void
wcfxs_post_initialize(struct wcfxs_softc *sc)
{
	int x;
	
	/* Finalize signalling */
	for (x = 0; x < sc->sc_cards; x++) {
	    if (sc->sc_cardmap & (1 << x)) {
		if (sc->sc_modtype[x] == MOD_TYPE_FXO)
		    sc->sc_chans[x].sigcap = ZT_SIG_FXSKS | ZT_SIG_FXSLS |
			ZT_SIG_SF;
		else
		    sc->sc_chans[x].sigcap = ZT_SIG_FXOKS | ZT_SIG_FXOLS |
			ZT_SIG_SF | ZT_SIG_EM;
	    }
	}
}

static int
wcfxs_hardware_init(struct wcfxs_softc *sc)
{
	unsigned char ver,x,y,c;
	u_int32_t write_dma_addr,read_dma_addr;
	int failed;

	/* I'd love to get rid of some of these magic numbers - riz */

	/* Signal a reset */
	WCFXS_WRITE1(WC_CNTL, 0x01);

	/* Check Freshmaker chip */
	x = WCFXS_READ1(WC_CNTL);
	ver = __wcfxs_getcreg(sc, WC_VER);
	failed = 0;
	if (ver != 0x59) {
	    printf("Freshmaker version: %02x\n", ver);
	    for (x = 0; x < 255; x++) {
		/* Test registers */
		if (ver >= 0x70) {
		    __wcfxs_setcreg(sc, WC_CS, x);
		    y = __wcfxs_getcreg(sc, WC_CS);
		} else {
		    __wcfxs_setcreg(sc, WC_TEST, x);
		    y = __wcfxs_getcreg(sc, WC_TEST);
		}
		if (x != y) {
		    printf("%02x != %02x\n", x, y);
		    failed++;
		}
	    }
	    if (!failed)
		printf("Freshmaker passed register test\n");
	    else {
		printf("Freshmaker failed register test\n");
		return -1;
	    }

	    /* Go to half-duty FSYNC */
	    __wcfxs_setcreg(sc, WC_SYNC, 0x01);
	    y = __wcfxs_getcreg(sc, WC_SYNC);
	} else {
	    printf("No freshmaker chip\n");
	}

	/* Reset PCI Interface chip and registers (and serial) */
	WCFXS_WRITE1(WC_CNTL, 0x06);
	/* Set up our proper outputs for when we switch to "serial" port */
	sc->sc_ios = BIT_CS | BIT_SCLK | BIT_SDI;

	WCFXS_WRITE1(WC_AUXD, sc->sc_ios);

	/* Set all to outputs except AUX 5, which is an input */
	WCFXS_WRITE1(WC_AUXC, ~AUX5_IN);

	/* Wait 1/4 sec */
	tsleep(sc, PZERO, "wcfxs_hardware_init", (hz / 4) + 1);

	/* Back to normal, with automatic DMA wraparound */
	WCFXS_WRITE1(WC_CNTL, 0x30 | 0x01);

	/* Make sure serial port and DMA are out of reset */
	c = WCFXS_READ1(WC_CNTL);
	c &= 0xf9;
	WCFXS_WRITE1(WC_CNTL, c);

	/* Configure serial port for MSB->LSB operation */
	WCFXS_WRITE1(WC_SERCTL, (SERIN_MSB_FIRST | SEROUT_MSB_FIRST | SER2XDCLK));

	/* Delay FSC by 0 so it's properly aligned */
	WCFXS_WRITE1(WC_FSCDELAY, 0x0);

	/* Setup DMA Addresses - XXX riz */
	write_dma_addr = vtophys((vaddr_t)sc->sc_write.chunk);
	WCFXS_WRITE4(WC_DMAWS, (unsigned int)write_dma_addr); /* write start */
	WCFXS_WRITE4(WC_DMAWI, (unsigned int)write_dma_addr +
	    ZT_CHUNKSIZE * 4 - 4);	/* Middle (interrupt) */
	WCFXS_WRITE4(WC_DMAWE, (unsigned int)write_dma_addr +
	    ZT_CHUNKSIZE * 8 - 4);	/* End */

	read_dma_addr = vtophys((vaddr_t)sc->sc_read.chunk);
	WCFXS_WRITE4(WC_DMARS, (unsigned int)read_dma_addr); /* write start */
	WCFXS_WRITE4(WC_DMARI, (unsigned int)read_dma_addr +
	    ZT_CHUNKSIZE * 4 - 4);	/* Middle (interrupt) */
	WCFXS_WRITE4(WC_DMARE, (unsigned int)read_dma_addr +
	    ZT_CHUNKSIZE * 8 - 4);	/* End */

	/* Clear interrupts */
	WCFXS_WRITE1(WC_INTSTAT, 0xff);

	/* wait another 1/2 sec */
	tsleep(sc, PZERO, "wcfxs_hardware_init", (hz / 2) + 1);

	for (x = 0; x < sc->sc_cards; x++) {
	    int sane = 0, ret = 0;

	    /* Init with Auto Calibration */
	    if (!(ret = wcfxs_init_proslic(sc, x, 0, 0, sane))) {
		sc->sc_cardmap |= (1 << x);
		printf("Module %d: Installed -- AUTO FXS\n", x);
	    } else {
		if (ret != -2) {
		    sane = 1;
		    /* Init with manual calibration */
		    if (!wcfxs_init_proslic(sc, x, 0, 1, sane)) {
			sc->sc_cardmap |= (1 << x);
			printf("Module %d: Installed -- MANUAL FXS\n", x);
		    } else {
			printf("Module %d: FAILED FXS\n", x);
		    }
		} else if (!(ret = wcfxs_init_voicedaa(sc, x, 0, 0, sane))) {
		    sc->sc_cardmap |= (1 << x);
		    printf("Module %d: Installed -- AUTO FXO\n", x);
		} else
		    printf("Module %d: Not installed\n", x);
	    }
	}

	/* return an error if no modules initialized correctly */
	if (!sc->sc_cardmap)
	    return -1;
	__wcfxs_setcreg(sc, WC_SYNC, (sc->sc_cardmap << 1) | 0x1);
	return 0;
}

static void wcfxs_enable_interrupts(struct wcfxs_softc *sc)
{
	/* Enable interrupts (we care about all of them) */
	WCFXS_WRITE1(WC_MASK0, 0x3f);
	/* No external interrupts */
	WCFXS_WRITE1(WC_MASK1, 0x00);
}

static void wcfxs_restart_dma(struct wcfxs_softc *sc)
{
	/* Reset Master and TDM */
	WCFXS_WRITE1(WC_CNTL, 0x01);
	WCFXS_WRITE1(WC_OPER, 0x01);
}

static void wcfxs_start_dma(struct wcfxs_softc *sc)
{
	/* Reset Master and TDM */
	WCFXS_WRITE1(WC_CNTL, 0x0f);
	tsleep(sc, PZERO | PCATCH, "startdma", 1);
	WCFXS_WRITE1(WC_CNTL, 0x01);
	WCFXS_WRITE1(WC_OPER, 0x01);
}

static void wcfxs_stop_dma(struct wcfxs_softc *sc)
{
	WCFXS_WRITE1(WC_OPER, 0x00);
}

static void wcfxs_reset_tdm(struct wcfxs_softc *sc)
{
	/* Reset TDM */
	WCFXS_WRITE1(WC_CNTL, 0x0f);
}

static void wcfxs_disable_interrupts(struct wcfxs_softc *sc)	
{
	WCFXS_WRITE1(WC_MASK0, 0x00);
	WCFXS_WRITE1(WC_MASK1, 0x00);
}

static __inline void wcfxs_transmitprep(struct wcfxs_softc *sc, unsigned char ints)
{
	volatile unsigned int *writechunk;
	int x;
	if (ints & 0x01) 
		/* Write is at interrupt address.  Start writing from normal offset */
		writechunk = sc->sc_write.chunk;
	else 
		writechunk = sc->sc_write.chunk + ZT_CHUNKSIZE;
	/* Calculate Transmission */
	zt_transmit(&sc->sc_span);

	for (x=0;x<ZT_CHUNKSIZE;x++) {
		/* Send a sample, as a 32-bit word */
		writechunk[x] = 0;
		if (sc->sc_cardmap & (1 << 3))
			writechunk[x] |= (sc->sc_chans[3].writechunk[x] << 24);
		if (sc->sc_cardmap & (1 << 2))
			writechunk[x] |= (sc->sc_chans[2].writechunk[x] << 16);
		if (sc->sc_cardmap & (1 << 1))
			writechunk[x] |= (sc->sc_chans[1].writechunk[x] << 8);
		if (sc->sc_cardmap & (1 << 0))
			writechunk[x] |= (sc->sc_chans[0].writechunk[x]);
		
	}

}

#ifdef AUDIO_RINGCHECK
static __inline void ring_check(struct wcfxs_softc *sc, int card)
{
	int x;
	short sample;
	if (sc->sc_modtype[card] != MOD_TYPE_FXO)
		return;
	sc->sc_mod.fxo.pegtimer[card] += ZT_CHUNKSIZE;
	for (x=0;x<ZT_CHUNKSIZE;x++) {
		/* Look for pegging to indicate ringing */
		sample = ZT_XLAW(sc->sc_chans[card].readchunk[x], (&(sc->sc_chans[card])));
		if ((sample > 10000) && (sc->sc_mod.fxo.peg[card] != 1)) {
			/* printf("High peg!\n"); */
			if ((sc->sc_mod.fxo.pegtimer[card] < PEGTIME) && (sc->sc_mod.fxo.pegtimer[card] > MINPEGTIME))
				sc->sc_mod.fxo.pegcount[card]++;
			sc->sc_mod.fxo.pegtimer[card] = 0;
			sc->sc_mod.fxo.peg[card] = 1;
		} else if ((sample < -10000) && (sc->sc_mod.fxo.peg[card] != -1)) {
			/* printf("Low peg!\n"); */
			if ((sc->sc_mod.fxo.pegtimer[card] < (PEGTIME >> 2)) && (sc->sc_mod.fxo.pegtimer[card] > (MINPEGTIME >> 2)))
				sc->sc_mod.fxo.pegcount[card]++;
			sc->sc_mod.fxo.pegtimer[card] = 0;
			sc->sc_mod.fxo.peg[card] = -1;
		}
	}
	if (sc->sc_mod.fxo.pegtimer[card] > PEGTIME) {
		/* Reset pegcount if our timer expires */
		sc->sc_mod.fxo.pegcount[card] = 0;
	}
	/* Decrement debouncer if appropriate */
	if (sc->sc_mod.fxo.ringdebounce[card])
		sc->sc_mod.fxo.ringdebounce[card]--;
	if (!sc->sc_mod.fxo.offhook[card] && !sc->sc_mod.fxo.ringdebounce[card]) {
		if (!sc->sc_mod.fxo.ring[card] && (sc->sc_mod.fxo.pegcount[card] > PEGCOUNT)) {
			/* It's ringing */
			DPRINTF(1,("RING on %d/%d!\n", sc->sc_span.spanno, card + 1));
			zt_hooksig(&sc->sc_chans[card], ZT_RXSIG_RING);
			if (!sc->sc_mod.fxo.offhook[card])
				zt_hooksig(&sc->sc_chans[card], ZT_RXSIG_RING);
			sc->sc_mod.fxo.ring[card] = 1;
		}
		if (sc->sc_mod.fxo.ring[card] && !sc->sc_mod.fxo.pegcount[card]) {
			/* No more ring */
			DPRINTF(1,("NO RING on %d/%d!\n", sc->sc_span.spanno, card + 1));
			zt_hooksig(&sc->sc_chans[card], ZT_RXSIG_OFFHOOK);
			sc->sc_mod.fxo.ring[card] = 0;
		}
	}
}
#endif


void wcfxs_generate_description(struct wcfxs_softc * sc)
{
	int x;
	/* prepare description */

	/* 80(name) + 40(description) + 25 per port description 
	 * + 20 just in case
	 */
	printf("generate_desc, cards = %d \n",sc->sc_cards);
	if(!card_configuration)
		card_configuration = malloc(80 + 40 +  sc->sc_cards*25 + 20, 
						M_TDM, M_NOWAIT);
	snprintf(card_configuration, 120 + sc->sc_cards*25, "%s\n%s\n", 
					sc->sc_span.name, sc->sc_span.desc);
	for( x = 0; x < sc->sc_cards; x++)
	{
			if(sc->sc_cardmap & (1 << x))
			{

				if(sc->sc_chans[x].sigcap & ZT_SIG_FXSKS) 	
					snprintf(card_configuration, 120 + sc->sc_cards*25, 
									"%sPort #%d -- FXO [%s]\n",
									card_configuration, x+1, sc->sc_mod.fxo.battery[x] ? "on" : "off");
				else
					snprintf(card_configuration, 120 + sc->sc_cards*25, 
									"%sPort #%d -- FXS\n",
									card_configuration, x+1);
			} else
			{
					snprintf(card_configuration, 120 + sc->sc_cards*25, 
									"%sPort #%d -- None\n",
									card_configuration, x+1);
			}
	}
}


static __inline void wcfxs_receiveprep(struct wcfxs_softc *sc, unsigned char ints)
{
	volatile unsigned int *readchunk;
	int x;

	if (ints & 0x08)
		readchunk = sc->sc_read.chunk + ZT_CHUNKSIZE;
	else
		/* Read is at interrupt address.  Valid data is available at normal offset */
		readchunk = sc->sc_read.chunk;
	for (x=0;x<ZT_CHUNKSIZE;x++) {
		if (sc->sc_cardmap & (1 << 3))
			sc->sc_chans[3].readchunk[x] = (readchunk[x] >> 24) & 0xff;
		if (sc->sc_cardmap & (1 << 2))
			sc->sc_chans[2].readchunk[x] = (readchunk[x] >> 16) & 0xff;
		if (sc->sc_cardmap & (1 << 1))
			sc->sc_chans[1].readchunk[x] = (readchunk[x] >> 8) & 0xff;
		if (sc->sc_cardmap & (1 << 0))
			sc->sc_chans[0].readchunk[x] = (readchunk[x]) & 0xff;
	}

#ifdef AUDIO_RINGCHECK
	for (x=0;x<sc->sc_cards;x++)
		ring_check(sc, x);
#endif		

	/* XXX We're wasting 8 taps.  We should get closer :( */
	for (x=0;x<sc->sc_cards;x++) {
		if (sc->sc_cardmap & (1 << x))
			zt_ec_chunk(&sc->sc_chans[x], sc->sc_chans[x].readchunk, sc->sc_chans[x].writechunk);
	}
	zt_receive(&sc->sc_span);
}

static __inline void __write_8bits(struct wcfxs_softc *sc, unsigned char bits)
{
	/* Drop chip select */
	int x;
	sc->sc_ios |= BIT_SCLK;
	WCFXS_WRITE1(WC_AUXD, sc->sc_ios);
	sc->sc_ios &= ~BIT_CS;
	WCFXS_WRITE1(WC_AUXD, sc->sc_ios);
	for (x=0;x<8;x++) {
		/* Send out each bit, MSB first, drop SCLK as we do so */
		if (bits & 0x80)
			sc->sc_ios |= BIT_SDI;
		else
			sc->sc_ios &= ~BIT_SDI;
		sc->sc_ios &= ~BIT_SCLK;
		WCFXS_WRITE1(WC_AUXD, sc->sc_ios);
		/* Now raise SCLK high again and repeat */
		sc->sc_ios |= BIT_SCLK;
		WCFXS_WRITE1(WC_AUXD, sc->sc_ios);
		bits <<= 1;
	}
	/* Finally raise CS back high again */
	sc->sc_ios |= BIT_CS;
	WCFXS_WRITE1(WC_AUXD, sc->sc_ios);
	
}


static __inline void __reset_spi(struct wcfxs_softc *sc)
{
	/* Drop chip select and clock once and raise and clock once */
	sc->sc_ios |= BIT_SCLK;
	WCFXS_WRITE1(WC_AUXD, sc->sc_ios);
	sc->sc_ios &= ~BIT_CS;
	WCFXS_WRITE1(WC_AUXD, sc->sc_ios);
	sc->sc_ios |= BIT_SDI;
	sc->sc_ios &= ~BIT_SCLK;
	WCFXS_WRITE1(WC_AUXD, sc->sc_ios);
	/* Now raise SCLK high again and repeat */
	sc->sc_ios |= BIT_SCLK;
	WCFXS_WRITE1(WC_AUXD, sc->sc_ios);
	/* Finally raise CS back high again */
	sc->sc_ios |= BIT_CS;
	WCFXS_WRITE1(WC_AUXD, sc->sc_ios);
	/* Clock again */
	sc->sc_ios &= ~BIT_SCLK;
	WCFXS_WRITE1(WC_AUXD, sc->sc_ios);
	/* Now raise SCLK high again and repeat */
	sc->sc_ios |= BIT_SCLK;
	WCFXS_WRITE1(WC_AUXD, sc->sc_ios);
	
}



static __inline unsigned char __read_8bits(struct wcfxs_softc *sc)
{
	unsigned char res=0, c;
	int x;
	sc->sc_ios |= BIT_SCLK;
	WCFXS_WRITE1(WC_AUXD, sc->sc_ios);
	/* Drop chip select */
	sc->sc_ios &= ~BIT_CS;
	WCFXS_WRITE1(WC_AUXD, sc->sc_ios);
	for (x=0;x<8;x++) {
		res <<= 1;
		/* Get SCLK */
		sc->sc_ios &= ~BIT_SCLK;
		WCFXS_WRITE1(WC_AUXD, sc->sc_ios);
		/* Read back the value */
		c = WCFXS_READ1(WC_AUXR);
		if (c & BIT_SDO)
			res |= 1;
		/* Now raise SCLK high again */
		sc->sc_ios |= BIT_SCLK;
		WCFXS_WRITE1(WC_AUXD, sc->sc_ios);
	}
	/* Finally raise CS back high again */
	sc->sc_ios |= BIT_CS;
	WCFXS_WRITE1(WC_AUXD, sc->sc_ios);
	sc->sc_ios &= ~BIT_SCLK;
	WCFXS_WRITE1(WC_AUXD, sc->sc_ios);

	/* And return our result */
	return res;
}

static void __wcfxs_setcreg(struct wcfxs_softc *sc, unsigned char reg, unsigned char val)
{
	WCFXS_WRITE1(WC_REGBASE + ((reg & 0xf) << 2), val);
}

static unsigned char __wcfxs_getcreg(struct wcfxs_softc *sc, unsigned char reg)
{
	return WCFXS_READ1(WC_REGBASE + ((reg & 0xf) << 2));
}

static __inline void __wcfxs_setcard(struct wcfxs_softc *sc, int card)
{
	if (sc->sc_curcard != card) {
		__wcfxs_setcreg(sc, WC_CS, (1 << card));
		sc->sc_curcard = card;
	}
}

static void __wcfxs_setreg(struct wcfxs_softc *sc, int card, unsigned char reg, unsigned char value)
{
	__wcfxs_setcard(sc, card);
	if (sc->sc_modtype[card] == MOD_TYPE_FXO) {
		__write_8bits(sc, 0x20);
		__write_8bits(sc, reg & 0x7f);
	} else {
		__write_8bits(sc, reg & 0x7f);
	}
 
	__write_8bits(sc, value);
}

static void wcfxs_setreg(struct wcfxs_softc *sc, int card, unsigned char reg, unsigned char value)
{
	FXS_LOCK(sc->sc_lock);
	__wcfxs_setreg(sc, card, reg, value);
	FXS_UNLOCK(sc->sc_lock);
}

static unsigned char __wcfxs_getreg(struct wcfxs_softc *sc, int card, unsigned char reg)
{
	__wcfxs_setcard(sc, card);
	if (sc->sc_modtype[card] == MOD_TYPE_FXO) {
		__write_8bits(sc, 0x60);
		__write_8bits(sc, reg & 0x7f);
	} else {
		__write_8bits(sc, reg | 0x80);
	}

	return __read_8bits(sc);
}

static __inline void reset_spi(struct wcfxs_softc *sc, int card)
{
	FXS_LOCK(sc->sc_lock);
	__wcfxs_setcard(sc, card);
	__reset_spi(sc);
	__reset_spi(sc);
	FXS_UNLOCK(sc->sc_lock);
}

static unsigned char wcfxs_getreg(struct wcfxs_softc *sc, int card, unsigned char reg)
{
	unsigned char res;
	FXS_LOCK(sc->sc_lock);
	res = __wcfxs_getreg(sc, card, reg);
	FXS_UNLOCK(sc->sc_lock);
	return res;
}

static int __wait_access(struct wcfxs_softc *sc, int card)
{
    unsigned char data;
    int count = 0;

    #define MAX_ATTEMPTS 6000 /* attempts */


    /* Wait for indirect access */
    while (count++ < MAX_ATTEMPTS)
	 {
		data = __wcfxs_getreg(sc, card, I_STATUS);

		if (!data)
			return 0;

	 }

    if(count > (MAX_ATTEMPTS-1)) printf(" ##### Loop error (%02x) #####\n", data);

	return 0;
}

static int wcfxs_proslic_setreg_indirect(struct wcfxs_softc *sc, int card, unsigned char address, unsigned short data)
{
	int res = -1;
	FXS_LOCK(sc->sc_lock);
	if(!__wait_access(sc, card)) {
		__wcfxs_setreg(sc, card, IDA_LO,(unsigned char)(data & 0xFF));
		__wcfxs_setreg(sc, card, IDA_HI,(unsigned char)((data & 0xFF00)>>8));
		__wcfxs_setreg(sc, card, IAA,address);
		res = 0;
	};
	FXS_UNLOCK(sc->sc_lock);
	return res;
}

static int wcfxs_proslic_getreg_indirect(struct wcfxs_softc *sc, int card, unsigned char address)
{ 
	int res = -1;
	char *p=NULL;
	FXS_LOCK(sc->sc_lock);
	if (!__wait_access(sc, card)) {
		__wcfxs_setreg(sc, card, IAA, address);
		if (!__wait_access(sc, card)) {
			unsigned char data1, data2;
			data1 = __wcfxs_getreg(sc, card, IDA_LO);
			data2 = __wcfxs_getreg(sc, card, IDA_HI);
			res = data1 | (data2 << 8);
		} else
			p = "Failed to wait inside\n";
	} else
		p = "failed to wait\n";
	FXS_UNLOCK(sc->sc_lock);
	if (p)
		printf(p);
	return res;
}

static int wcfxs_proslic_init_indirect_regs(struct wcfxs_softc *sc, int card)
{
	unsigned char i;

	for (i=0; i<sizeof(indirect_regs) / sizeof(indirect_regs[0]); i++)
	{
		if(wcfxs_proslic_setreg_indirect(sc, card, indirect_regs[i].address,indirect_regs[i].initial))
			return -1;
	}

	return 0;
}

static int wcfxs_proslic_verify_indirect_regs(struct wcfxs_softc *sc, int card)
{ 
	int passed = 1;
	unsigned short i, initial;
	int j;

	for (i=0; i<sizeof(indirect_regs) / sizeof(indirect_regs[0]); i++) 
	{
		if((j = wcfxs_proslic_getreg_indirect(sc, card, (unsigned char) indirect_regs[i].address)) < 0) {
			printf("Failed to read indirect register %d\n", i);
			return -1;
		}
		initial= indirect_regs[i].initial;

		if ( j != initial )
		{
			 printf("!!!!!!! %s  iREG %X = %X  should be %X\n",
				indirect_regs[i].name,indirect_regs[i].address,j,initial );
			 passed = 0;
		}	
	}

    if (passed) {
		DPRINTF(1,("Init Indirect Registers completed successfully.\n"));
    } else {
		printf(" !!!!! Init Indirect Registers UNSUCCESSFULLY.\n");
		return -1;
    }
    return 0;
}

static int wcfxs_voicedaa_insane(struct wcfxs_softc *sc, int card)
{
	int blah;
	blah = wcfxs_getreg(sc, card, 2);
	if (blah != 0x3)
		return -2;
	blah = wcfxs_getreg(sc, card, 11);
	DPRINTF(1,("VoiceDAA System: %02x\n", blah & 0xf));
	return 0;
}

static int wcfxs_proslic_insane(struct wcfxs_softc *sc, int card)
{
	int blah,insane_report;
	insane_report=0;

	blah = wcfxs_getreg(sc, card, 0);
	DPRINTF(1,("ProSLIC on module %d, product %d, version %d\n", card,
			   (blah & 0x30) >> 4, (blah & 0xf)));

#if	0
	if ((blah & 0x30) >> 4) {
		printf("ProSLIC on module %d is not a 3210.\n", card);
		return -1;
	}
#endif

	if (((blah & 0xf) == 0) || ((blah & 0xf) == 0xf)) {
		/* SLIC not loaded */
		return -1;
	}
	if ((blah & 0xf) < 3) {
		printf("ProSLIC 3210 version %d is too old\n", blah & 0xf);
		return -1;
	}

	blah = wcfxs_getreg(sc, card, 8);
	if (blah != 0x2) {
		printf("ProSLIC on module %d insane (1) %d should be 2\n", card, blah);
		return -1;
	} else if ( insane_report)
		printf("ProSLIC on module %d Reg 8 Reads %d Expected is 0x2\n",card,blah);

	blah = wcfxs_getreg(sc, card, 64);
	if (blah != 0x0) {
		printf("ProSLIC on module %d insane (2)\n", card);
		return -1;
	} else if ( insane_report)
		printf("ProSLIC on module %d Reg 64 Reads %d Expected is 0x0\n",card,blah);

	blah = wcfxs_getreg(sc, card, 11);
	if (blah != 0x33) {
		printf("ProSLIC on module %d insane (3)\n", card);
		return -1;
	} else if ( insane_report)
		printf("ProSLIC on module %d Reg 11 Reads %d Expected is 0x33\n",card,blah);

	/* Just be sure it's setup right. */
	wcfxs_setreg(sc, card, 30, 0);

	DPRINTF(1,("ProSLIC on module %d seems sane.\n", card));
	return 0;
}

static int wcfxs_proslic_powerleak_test(struct wcfxs_softc *sc, int card)
{
	unsigned long origjiffies;
	unsigned long jiffies;
	unsigned char vbat;
	struct timeval tv;

	/* Turn off linefeed */
	wcfxs_setreg(sc, card, 64, 0);

	/* Power down */
	wcfxs_setreg(sc, card, 14, 0x10);

	/* Wait for one second */
	microtime(&tv);
	origjiffies = tvtohz(&tv);
	jiffies = origjiffies;

	while((vbat = wcfxs_getreg(sc, card, 82)) > 0x6) {
		microtime(&tv);
		jiffies = tvtohz(&tv);
		if ((jiffies - origjiffies) >= (hz/2))
			break;;
	}

	if (vbat < 0x06) {
		printf("Excessive leakage detected on module %d: %d volts (%02x) after %d ms\n", card,
		       376 * vbat / 1000, vbat, (int)((jiffies - origjiffies) * 1000 / hz));
		return -1;
	} else 
		DPRINTF(1,("Post-leakage voltage: %d volts\n", 376 * vbat / 1000));
	return 0;
}

static int wcfxs_powerup_proslic(struct wcfxs_softc *sc, int card, int fast)
{
	unsigned char vbat;
	unsigned long origjiffies;
	unsigned long jiffies;
	struct timeval tv;

	/* Set period of DC-DC converter to 1/64 khz */
	wcfxs_setreg(sc, card, 92, 0xff /* was 0xff */);

	/* Engage DC-DC converter */
	/* wcfxs_setreg(sc, card, 93, 0x99); */

	/* Wait for VBat to powerup */
	microtime(&tv);
	jiffies = tvtohz(&tv);
	origjiffies = jiffies;

	/* Disable powerdown */
	wcfxs_setreg(sc, card, 14, 0);

	/* If fast, don't bother checking anymore */
	if (fast)
		return 0;

	while((vbat = wcfxs_getreg(sc, card, 82)) < 0xc0) {
		/* Wait no more than 500ms */
		microtime(&tv);
		jiffies = tvtohz(&tv);
		if ((jiffies - origjiffies) > hz/2) {
			break;
		}
	}

	if (vbat < 0xc0) {
		printf("ProSLIC on module %d failed to powerup within %d ms (%d mV only)\n",
		       card, (int)(((jiffies - origjiffies) * 1000 / hz)),
			vbat * 375);
		return -1;
	} else
		DPRINTF(1,("ProSLIC on module %d powered up to -%d volts "
				   "(%02x) in %d ms\n", card, vbat * 376 / 1000,
				   vbat, (int)(((jiffies - origjiffies) * 1000 / hz))));

	/* Perform DC-DC calibration */
	/* wcfxs_setreg(sc, card, 93, 0x80); */
	/* Engage DC-DC converter */
	wcfxs_setreg(sc, card, 93, 0x99 /* was 0x19 */);

#if 0
	origjiffies = jiffies;
	while(0x80 & wcfxs_getreg(sc, card, 93)) {
		if ((jiffies - origjiffies) > 2 * HZ) {
			printf("Timeout waiting for DC-DC calibration on module %d\n", card);
			return -1;
		}
	}

#if 0
	/* Wait a full two seconds */
	while((jiffies - origjiffies) < 2 * HZ);

	/* Just check to be sure */
	vbat = wcfxs_getreg(sc, card, 82);
	printf("ProSLIC on module %d powered up to -%d volts (%02x) in %d ms\n",
		       card, vbat * 376 / 1000, vbat, (int)(((jiffies - origjiffies) * 1000 / HZ)));
#endif
#endif
	return 0;

}

static int wcfxs_proslic_manual_calibrate(struct wcfxs_softc *sc, int card){
	unsigned long origjiffies;
	unsigned long jiffies;
	unsigned char i;
	struct timeval tv;

	wcfxs_setreg(sc, card, 21, 0);/* (0)  Disable all interupts in DR21 */
	wcfxs_setreg(sc, card, 22, 0);/* (0)Disable all interupts in DR21 */
	wcfxs_setreg(sc, card, 23, 0);/* (0)Disable all interupts in DR21 */
	wcfxs_setreg(sc, card, 64, 0);/* (0) */

	wcfxs_setreg(sc, card, 97, 0x18); /* (0x18)Calibrations without the ADC and DAC offset and without common mode calibration. */
	wcfxs_setreg(sc, card, 96, 0x47); /* (0x47)	Calibrate common mode and differential DAC mode DAC + ILIM */

	microtime(&tv);
	jiffies = tvtohz(&tv);
	origjiffies=jiffies;
	while( wcfxs_getreg(sc,card,96)!=0 ){
		microtime(&tv);
		jiffies = tvtohz(&tv);
		if((jiffies-origjiffies)>80*hz/100)
			return -1;
	}
/*Initialized DR 98 and 99 to get consistant results.
 *  98 and 99 are the results registers and the search 
 *  should have same intial conditions. 
 */

/*******************************The following is the manual gain mismatch calibration****************************/
/*******************************This is also available as a function *******************************************/
	/* Delay 10ms */
	origjiffies=jiffies; 
	while((jiffies-origjiffies)<1*hz/100)
	{
		microtime(&tv);
		jiffies = tvtohz(&tv);
	}

	wcfxs_proslic_setreg_indirect(sc, card, 88, 0);
	wcfxs_proslic_setreg_indirect(sc, card, 89, 0);
	wcfxs_proslic_setreg_indirect(sc, card, 90, 0);
	wcfxs_proslic_setreg_indirect(sc, card, 91, 0);
	wcfxs_proslic_setreg_indirect(sc, card, 92, 0);
	wcfxs_proslic_setreg_indirect(sc, card, 93, 0);
 

	wcfxs_setreg(sc, card, 98,0x10); /*  This is necessary if the calibration occurs other than at reset time */
	wcfxs_setreg(sc, card, 99,0x10);

	for ( i=0x1f; i>0; i--)
	{
		wcfxs_setreg(sc, card, 98,i);
		microtime(&tv);
		jiffies = tvtohz(&tv);
		origjiffies=jiffies; 

		while((jiffies-origjiffies)<4*hz/100)
		{
			microtime(&tv);
			jiffies = tvtohz(&tv);
		}

		if((wcfxs_getreg(sc, card, 88)) == 0)
			break;
	} /*  for */

	for ( i=0x1f; i>0; i--)
	{
		wcfxs_setreg(sc, card, 99,i);
		microtime(&tv);
		jiffies = tvtohz(&tv);
		origjiffies=jiffies; 
		while((jiffies-origjiffies)<4*hz/100)
		{
			microtime(&tv);
			jiffies = tvtohz(&tv);
		};
		if((wcfxs_getreg(sc, card, 89)) == 0)
			break;
	}/* for */

/*******************************The preceding is the manual gain mismatch calibration****************************/
/**********************************The following is the longitudinal Balance Cal***********************************/
	wcfxs_setreg(sc, card, 64, 1);
	microtime(&tv);
	jiffies = tvtohz(&tv);
	origjiffies=jiffies; 
	/*  Sleep 100? */
	while((jiffies-origjiffies)<10*hz/100)
	{
		microtime(&tv);
		jiffies = tvtohz(&tv);
	}	

	wcfxs_setreg(sc, card, 64, 0);
	wcfxs_setreg(sc, card, 23, 0x4);  /*  enable interrupt for the balance Cal */
	wcfxs_setreg(sc, card, 97, 0x1); /*  this is a singular calibration bit for longitudinal calibration */
	wcfxs_setreg(sc, card, 96,0x40);

	wcfxs_getreg(sc, card, 96); /* Read Reg 96 just cause */

	wcfxs_setreg(sc, card, 21, 0xFF);
	wcfxs_setreg(sc, card, 22, 0xFF);
	wcfxs_setreg(sc, card, 23, 0xFF);

	/**The preceding is the longitudinal Balance Cal***/
	return(0);

}
#if 1
static int wcfxs_proslic_calibrate(struct wcfxs_softc *sc, int card)
{
	unsigned long origjiffies;
	unsigned long jiffies;
	struct timeval tv;
	int x;
	/* Perform all calibrations */
	wcfxs_setreg(sc, card, 97, 0x1f);
	
	/* Begin, no speedup */
	wcfxs_setreg(sc, card, 96, 0x5f);

	/* Wait for it to finish */
	microtime(&tv);
	jiffies = tvtohz(&tv);
	origjiffies = jiffies;

	while(wcfxs_getreg(sc, card, 96)) {
		if ((jiffies - origjiffies) > 2 * hz) {
			printf("Timeout waiting for calibration of module %d\n", card);
			return -1;
		}
		microtime(&tv);
		jiffies = tvtohz(&tv);
	}
	
	DEBUG_DO(1,
		/* Print calibration parameters */
		printf("Calibration Vector Regs 98 - 107: \n");
		for (x=98;x<108;x++) {
			printf("%d: %02x\n", x, wcfxs_getreg(sc, card, x));
		}
		);
	return 0;
}
#endif

static int wcfxs_init_voicedaa(struct wcfxs_softc *sc, int card, int fast, int manual, int sane)
{
	unsigned long newjiffies;
	unsigned long jiffies;
	struct timeval tv;

	sc->sc_modtype[card] = MOD_TYPE_FXO;
	/* Sanity check the ProSLIC */
	reset_spi(sc, card);
	if (!sane && wcfxs_voicedaa_insane(sc, card))
		return -2;

	/* Software reset */
	wcfxs_setreg(sc, card, 1, 0x80);

	/* Wait just a bit */
	tsleep(sc, PZERO, "wcfxs_init_voicedaa", (hz / 10) + 1);

	/* Enable PCM, ulaw */
	wcfxs_setreg(sc, card, 33, 0x28);

	/* Misc. DAA parameters */
	wcfxs_setreg(sc, card, 31, 0xa3);

	/* Set AC Impedence */
	wcfxs_setreg(sc, card, 30, 0x00);

	/* Set DC Termination */
	wcfxs_setreg(sc, card, 26, 0xc0);

	/* Set Transmit/Receive timeslot */
	wcfxs_setreg(sc, card, 34, (3-card) * 8);
	wcfxs_setreg(sc, card, 35, 0x00);
	wcfxs_setreg(sc, card, 36, (3-card) * 8);
	wcfxs_setreg(sc, card, 37, 0x00);

	/* Enable ISO-Cap */
	wcfxs_setreg(sc, card, 6, 0x00);

	/* Wait 1000ms for ISO-cap to come up */

	microtime(&tv);
	jiffies = tvtohz(&tv);
	newjiffies = jiffies;
	newjiffies += 2 * hz;

	while((jiffies < newjiffies) && !(wcfxs_getreg(sc, card, 11) & 0xf0))
	{
		microtime(&tv);
		jiffies = tvtohz(&tv);
	}

	if (!(wcfxs_getreg(sc, card, 11) & 0xf0)) {
		printf("VoiceDAA did not bring up ISO link properly!\n");
		return -1;
	}
	DPRINTF(1,("ISO-Cap is now up, line side: %02x rev %02x\n", 
		       wcfxs_getreg(sc, card, 11) >> 4,
		       (wcfxs_getreg(sc, card, 13) >> 2) & 0xf));
	/* Enable on-hook line monitor */
	wcfxs_setreg(sc, card, 5, 0x08);
	return 0;
		
}

static int wcfxs_init_proslic(struct wcfxs_softc *sc, int card, int fast, int manual, int sane)
{

	unsigned short tmp[5];
	int x;

	/* By default, don't send on hook */
	sc->sc_mod.fxs.idletxhookstate [card] = 1;

	/* Sanity check the ProSLIC */
	if (!sane && wcfxs_proslic_insane(sc, card))
		return -2;
		
	if (sane) {
		/* Make sure we turn off the DC->DC converter to prevent anything from blowing up */
		wcfxs_setreg(sc, card, 14, 0x10);
	}

	if (wcfxs_proslic_init_indirect_regs(sc, card)) {
		printf("Indirect Registers failed to initialize on module %d.\n", card);
		return -1;
	}

	/* Clear scratch pad area */
	wcfxs_proslic_setreg_indirect(sc, card, 97,0);

	/* Clear digital loopback */
	wcfxs_setreg(sc, card, 8, 0);

	/* Revision C optimization */
	wcfxs_setreg(sc, card, 108, 0xeb);

	/* Disable automatic VBat switching for safety to prevent
	   Q7 from accidently turning on and burning out. */
	wcfxs_setreg(sc, card, 67, 0x17);

	/* Turn off Q7 */
	wcfxs_setreg(sc, card, 66, 1);

	/* Flush ProSLIC digital filters by setting to clear, while
	   saving old values */
	for (x=0;x<5;x++) {
		tmp[x] = wcfxs_proslic_getreg_indirect(sc, card, x + 35);
		wcfxs_proslic_setreg_indirect(sc, card, x + 35, 0x8000);
	}

	/* Power up the DC-DC converter */
	if (wcfxs_powerup_proslic(sc, card, fast)) {
		printf("Unable to do INITIAL ProSLIC powerup on module %d\n", card);
		return -1;
	}

	if (!fast) {

		/* Check for power leaks */
		if (wcfxs_proslic_powerleak_test(sc, card)) {
			printf("ProSLIC module %d failed leakage test.  Check for short circuit\n", card);
		}
		/* Power up again */
		if (wcfxs_powerup_proslic(sc, card, fast)) {
			printf("Unable to do FINAL ProSLIC powerup on module %d\n", card);
			return -1;
		}
#ifndef NO_CALIBRATION
		/* Perform calibration */
		if(manual) {
			if (wcfxs_proslic_manual_calibrate(sc, card)) {
				/* printf("Proslic failed on Manual Calibration\n"); */
				if (wcfxs_proslic_manual_calibrate(sc, card)) {
					printf("Proslic Failed on Second Attempt to Calibrate Manually. (Try -DNO_CALIBRATION in Makefile)\n");
					return -1;
				}
				printf("Proslic Passed Manual Calibration on Second Attempt\n");
			}
		}
		else {
			if(wcfxs_proslic_calibrate(sc, card))  {
				/* printf("ProSlic died on Auto Calibration.\n"); */
				if (wcfxs_proslic_calibrate(sc, card)) {
					printf("Proslic Failed on Second Attempt to Auto Calibrate\n");
					return -1;
				}
				printf("Proslic Passed Auto Calibration on Second Attempt\n");
			}
		}
#endif

	}
	/* Calibration complete, restore original values */
	for (x=0;x<5;x++) {
		wcfxs_proslic_setreg_indirect(sc, card, x + 35, tmp[x]);
	}

	if (wcfxs_proslic_verify_indirect_regs(sc, card)) {
		printf("Indirect Registers failed verification.\n");
		return -1;
	}


#if 0
    /* Disable Auto Power Alarm Detect and other "features" */
    wcfxs_setreg(sc, card, 67, 0x0e);
    blah = wcfxs_getreg(sc, card, 67);
#endif

#if 0
    if (wcfxs_proslic_setreg_indirect(sc, card, 97, 0x0)) { /*  Stanley: for the bad recording fix */
		 printf(KERN_INFO "ProSlic IndirectReg Died.\n");
		 return -1;
	}
#endif

    wcfxs_setreg(sc, card, 1, 0x28);
 	/*  U-Law 8-bit interface */
    wcfxs_setreg(sc, card, 2, (3-card) * 8);    /*  Tx Start count low byte  0 */
    wcfxs_setreg(sc, card, 3, 0);    /*  Tx Start count high byte 0 */
    wcfxs_setreg(sc, card, 4, (3-card) * 8);    /*  Rx Start count low byte  0 */
    wcfxs_setreg(sc, card, 5, 0);    /*  Rx Start count high byte 0 */
    wcfxs_setreg(sc, card, 18, 0xff);     /*  clear all interrupt */
    wcfxs_setreg(sc, card, 19, 0xff);
    wcfxs_setreg(sc, card, 20, 0xff);
    wcfxs_setreg(sc, card, 73, 0x04);

#if 0
    wcfxs_setreg(sc, card, 21, 0x00); 	/*  enable interrupt */
    wcfxs_setreg(sc, card, 22, 0x02); 	/*  Loop detection interrupt */
    wcfxs_setreg(sc, card, 23, 0x01); 	/*  DTMF detection interrupt */
    wcfxs_setreg(sc, card, 72, 0x20);
#endif

#if 0
    /* Enable loopback */
    wcfxs_setreg(sc, card, 8, 0x2);
    wcfxs_setreg(sc, card, 14, 0x0);
    wcfxs_setreg(sc, card, 64, 0x0);
    wcfxs_setreg(sc, card, 1, 0x08);
#endif

#ifdef BOOST_RINGER
	/* Beef up Ringing voltage to 89V */
	if (wcfxs_proslic_setreg_indirect(sc, card, 21, 0x1d1))
			return -1;
#endif
	return 0;
}

static void wcfxs_proslic_recheck_sanity(struct wcfxs_softc *sc, int card)
{
	int res;
	/* Check loopback */
	res = wcfxs_getreg(sc, card, 8);
	if (res) {
		printf("Ouch, part reset, quickly restoring reality (%d)\n", card);
		wcfxs_init_proslic(sc, card, 1, 0, 1);
	} else {
		res = wcfxs_getreg(sc, card, 64);
		if (!res && (res != sc->sc_mod.fxs.lasttxhook[card])) {
			if (sc->sc_mod.fxs.palarms[card]++ < MAX_ALARMS) {
				printf("Power alarm on module %d, resetting!\n", card + 1);
				if (sc->sc_mod.fxs.lasttxhook[card] == 4)
					sc->sc_mod.fxs.lasttxhook[card] = 1;
				wcfxs_setreg(sc, card, 64, sc->sc_mod.fxs.lasttxhook[card]);
			} else {
				if (sc->sc_mod.fxs.palarms[card] == MAX_ALARMS)
					printf("Too many power alarms on card %d, NOT resetting!\n", card + 1);
			}
		}
	}
}

static void wcfxs_voicedaa_check_hook(struct wcfxs_softc *sc, int card)
{
	signed char b;
#ifndef AUDIO_RINGCHECK
	unsigned char res;
#endif

	int poopy = 0;
	/* Try to track issues that plague slot one FXO's */
	b = wcfxs_getreg(sc, card, 5);
	if ((b & 0x2) || !(b & 0x8)) {
		/* Not good -- don't look at anything else */
		DPRINTF(1,("Poopy (%02x) on card %d!\n", b, card + 1));
		poopy++;
	}
	b &= 0x9b;
	if (sc->sc_mod.fxo.offhook[card]) {
		if (b != 0x9)
			wcfxs_setreg(sc, card, 5, 0x9);
	} else {
		if (b != 0x8)
			wcfxs_setreg(sc, card, 5, 0x8);
	}
	if (poopy)
		return;


#ifndef AUDIO_RINGCHECK
	if (!sc->sc_mod.fxo.offhook[card]) {
		res = wcfxs_getreg(sc, card, 5);
		if ((res & 0x60) && sc->sc_mod.fxo.battery[card]) {
			sc->sc_mod.fxo.ringdebounce[card] += (ZT_CHUNKSIZE * 4);
			if (sc->sc_mod.fxo.ringdebounce[card] >= ZT_CHUNKSIZE * 64) {
				if (!sc->sc_mod.fxo.wasringing[card]) {
					sc->sc_mod.fxo.wasringing[card] = 1;
					zt_hooksig(&sc->sc_chans[card], ZT_RXSIG_RING);
					DPRINTF(1,("RING!\n"));
				}
				sc->sc_mod.fxo.ringdebounce[card] = ZT_CHUNKSIZE * 32;
			}
		} else {
			sc->sc_mod.fxo.ringdebounce[card] -= ZT_CHUNKSIZE;
			if (sc->sc_mod.fxo.ringdebounce[card] <= 0) {
				if (sc->sc_mod.fxo.wasringing[card]) {
					sc->sc_mod.fxo.wasringing[card] =0;
					zt_hooksig(&sc->sc_chans[card], ZT_RXSIG_OFFHOOK);
					DPRINTF(1,("NO RING!\n"));
				}
				sc->sc_mod.fxo.ringdebounce[card] = 0;
			}
				
		}
	}
#endif
	b = wcfxs_getreg(sc, card, 29);
#if 0
	{
		static int count = 0;
		if (!(count++ % 100)) {
			printf("Voltage: %d\n", b);
		}
	}
#endif	
	if (abs(b) < BATT_THRESH) {
		sc->sc_mod.fxo.nobatttimer[card]++;
#if 0
		if (sc->sc_mod.fxo.battery[card])
			printf("Battery loss: %d (%d debounce)\n", b, sc->sc_mod.fxo.battdebounce[card]);
#endif
		if (sc->sc_mod.fxo.battery[card] && !sc->sc_mod.fxo.battdebounce[card]) {
			DPRINTF(1,("[%d] NO BATTERY!\n", card));
			sc->sc_mod.fxo.battery[card] =  0;
			wcfxs_generate_description(sc);
#ifdef	JAPAN
			if ((!sc->ohdebounce) && sc->offhook) {
				zt_hooksig(&sc->sc_chans[card], ZT_RXSIG_ONHOOK);
				DPRINTF(1,("Signalled On Hook\n"));
#ifdef	ZERO_BATT_RING
				sc->onhook++;
#endif
			}
#else
			zt_hooksig(&sc->sc_chans[card], ZT_RXSIG_ONHOOK);
#endif
			sc->sc_mod.fxo.battdebounce[card] = BATT_DEBOUNCE;
		} else if (!sc->sc_mod.fxo.battery[card])
			sc->sc_mod.fxo.battdebounce[card] = BATT_DEBOUNCE;
	} else if (abs(b) > BATT_THRESH) {
		if (!sc->sc_mod.fxo.battery[card] && !sc->sc_mod.fxo.battdebounce[card]) {
			DPRINTF(1,("[%d] BATTERY!\n", card));
#ifdef	ZERO_BATT_RING
			if (sc->onhook) {
				sc->onhook = 0;
				zt_hooksig(&sc->sc_chans[card], ZT_RXSIG_OFFHOOK);
				DPRINTF(1,("Signalled Off Hook\n"));
			}
#else
			zt_hooksig(&sc->sc_chans[card], ZT_RXSIG_OFFHOOK);
#endif
			sc->sc_mod.fxo.battery[card] = 1;
			sc->sc_mod.fxo.nobatttimer[card] = 0;
			sc->sc_mod.fxo.battdebounce[card] = BATT_DEBOUNCE;
			wcfxs_generate_description(sc);
		} else if (sc->sc_mod.fxo.battery[card])
			sc->sc_mod.fxo.battdebounce[card] = BATT_DEBOUNCE;
	} else {
		/* It's something else... */
		sc->sc_mod.fxo.battdebounce[card] = BATT_DEBOUNCE;
	}
	if (sc->sc_mod.fxo.battdebounce[card])
		sc->sc_mod.fxo.battdebounce[card]--;
	
}

static void wcfxs_proslic_check_hook(struct wcfxs_softc *sc, int card)
{
	char res;
	int hook;

	/* For some reason we have to debounce the
	   hook detector.  */

	res = wcfxs_getreg(sc, card, 68);
	hook = (res & 1);
	if (hook != sc->sc_mod.fxs.lastrxhook[card]) {
		/* Reset the debounce (must be multiple of 4ms) */
		sc->sc_mod.fxs.debounce[card] = 8 * (4 * 8);
#if 0
		printf("Resetting debounce card %d hook %d, %d\n", card, hook, sc->sc_mod.fxs.debounce[card]);
#endif
	} else {
		if (sc->sc_mod.fxs.debounce[card] > 0) {
			sc->sc_mod.fxs.debounce[card]-= 4 * ZT_CHUNKSIZE;
#if 0
			printf("Sustaining hook %d, %d\n", hook, sc->sc_mod.fxs.debounce[card]);
#endif
			if (!sc->sc_mod.fxs.debounce[card]) {
#if 0
				printf("Counted down debounce, newhook: %d...\n", hook);
#endif
				sc->sc_mod.fxs.debouncehook[card] = hook;
			}
			if (!sc->sc_mod.fxs.oldrxhook[card] && sc->sc_mod.fxs.debouncehook[card]) {
				/* Off hook */
				DPRINTF(1,("wcfxs: Card %d Going off hook\n", card));
				zt_hooksig(&sc->sc_chans[card], ZT_RXSIG_OFFHOOK);
				if (robust)
					wcfxs_init_proslic(sc, card, 1, 0, 1);
				sc->sc_mod.fxs.oldrxhook[card] = 1;
			
			} else if (sc->sc_mod.fxs.oldrxhook[card] && !sc->sc_mod.fxs.debouncehook[card]) {
				/* On hook */
				DPRINTF(1,("wcfxs: Card %d Going on hook\n", card));
				zt_hooksig(&sc->sc_chans[card], ZT_RXSIG_ONHOOK);
				sc->sc_mod.fxs.oldrxhook[card] = 0;
			}
		}
	}
	sc->sc_mod.fxs.lastrxhook[card] = hook;

	
}

static void
wcfxs_attach(struct device *parent, struct device *self, void *aux)
{
	struct wcfxs_softc *sc = (struct wcfxs_softc *) self;
	struct pci_attach_args *pa = aux;
	pci_chipset_tag_t pc = pa->pa_pc;
	pcireg_t subdev;
	pci_intr_handle_t ih;
	const char *intrstr;
	struct wcfxs_ident *id;
	int error;

#ifdef _LKM
	malloc_type_attach(M_TDM);  /* XXX this  belongs elsewhere */
#endif

	/* read the subdevice information - XXX code duplication */
	subdev = pci_conf_read(pc, pa->pa_tag, PCI_SUBSYS_ID_REG);

	for (id = wcfxs_ident_table ; id->device_info != NULL; id++)
		if (id->vendor == PCI_VENDOR(pa->pa_id) &&
			id->product == PCI_PRODUCT(pa->pa_id) &&
			(id->subvendor == PCI_VENDOR(subdev) ||
			id->subvendor == (pci_vendor_id_t) -1))
			break;

	if (id->device_info == NULL) {
		printf("Card info not found\n");
		return;
	}

	printf(": %s\n",id->device_info->name);

	/*
	 * Make sure bus mastering, ioports, and memory mapping are enabled.
	 */
	pci_conf_write(pc, pa->pa_tag, PCI_COMMAND_STATUS_REG,
		pci_conf_read(pc, pa->pa_tag, PCI_COMMAND_STATUS_REG) |
		PCI_COMMAND_MASTER_ENABLE | PCI_COMMAND_IO_ENABLE |
		PCI_COMMAND_MEM_ENABLE);

	if (!(pci_conf_read(pc, pa->pa_tag, PCI_COMMAND_STATUS_REG) &
		PCI_COMMAND_IO_ENABLE)) {
		printf("%s: failed to enable I/O ports\n",
			sc->sc_dv.dv_xname);
		goto fail;
	}
	if (!(pci_conf_read(pc, pa->pa_tag, PCI_COMMAND_STATUS_REG) &
		PCI_COMMAND_MEM_ENABLE)) {
		printf("%s: failed to enable memory mapping\n",
			sc->sc_dv.dv_xname);
		goto fail;
	}

	if ((error = pci_mapreg_map(pa, WCFXS_PCI_BAR, PCI_MAPREG_TYPE_IO, 0,
		&sc->sc_iot, &sc->sc_ioh, &sc->sc_iob, &sc->sc_iosize)) != 0) {
	    aprint_error("%s: can't map i/o space, error = %d\n",
		sc->sc_dv.dv_xname, error);
	    goto fail;
	}

	sc->sc_pct = pc;
	sc->sc_dmat = pa->pa_dmat;
	sc->sc_curcard = -1;
	sc->sc_cards = 4;   /* XXX magic number? */
	sc->sc_pos = sc->sc_dv.dv_unit;
	sc->sc_variety = id->device_info->name;
	sc->sc_flags = id->device_info->flags;
	sc->sc_usecount = 0;

	zaptel_attach_mi(&sc->sc_dv);

	/* Map and establish the interrupt */
	if (pci_intr_map(pa, &ih)) {
	    aprint_error("%s: can't map interrupt\n", sc->sc_dv.dv_xname);
	    goto fail;
	}

	intrstr = pci_intr_string(pc, ih);
	sc->sc_ih = pci_intr_establish(pc, ih, IPL_NET, wcfxs_intr, sc);
	if (sc->sc_ih == NULL) {
	    aprint_error("%s: unable to establish interrupt",
		sc->sc_dv.dv_xname);
	    if (intrstr != NULL)
		aprint_normal(" at %s", intrstr);
	    aprint_normal("\n");
	    goto fail;
	}
	aprint_normal("%s: interrupting at %s\n",sc->sc_dv.dv_xname, intrstr);

	/*
	 * Set up DMA mappings
	 */
	if (!wcfxs_dma_malloc(sc, &sc->sc_read))
	    goto fail;
	if (!wcfxs_dma_malloc(sc, &sc->sc_write))
	    goto fail_1;

	if (wcfxs_initialize(sc)) {
	    unsigned char x;

	    /* set reset low */
	    x = WCFXS_READ1(WC_CNTL);
	    WCFXS_WRITE1(WC_CNTL, (~0x1) & x);
	    aprint_error("%s: unable to initialize FXS\n",
		sc->sc_dv.dv_xname);
	    goto fail_2;
	}

	if (wcfxs_hardware_init(sc)) {
		unsigned char x;

		/* set reset low */
		x = WCFXS_READ1(WC_CNTL);
		WCFXS_WRITE1(WC_CNTL, (~0x1) & x);

		goto fail_2;
	}

	wcfxs_post_initialize(sc);
	wcfxs_generate_description(sc);

	/* Initialize buffers to 0 */
	memset((void *) sc->sc_read.chunk, 0, WCFXS_SIZE);
	memset((void *) sc->sc_write.chunk, 0, WCFXS_SIZE);

	wcfxs_enable_interrupts(sc);
	wcfxs_start_dma(sc);

	return;

 fail_2:
	wcfxs_dma_free(sc, &sc->sc_write);
 fail_1:
	wcfxs_dma_free(sc, &sc->sc_read);
 fail:
	return;
}

static int
wcfxs_detach(struct device* self, int flags)
{
	struct wcfxs_softc *sc = (struct wcfxs_softc *) self;

	if (sc) {
		/* don't unload if module is in use */
		if (sc->sc_usecount)
			return (EBUSY);
	}

	wcfxs_disable_interrupts(sc);
	if (sc->sc_ih != NULL)
	    pci_intr_disestablish(sc->sc_pct, sc->sc_ih);
	wcfxs_stop_dma(sc);
	wcfxs_reset_tdm(sc);

	/* free read/write dma buffers */
	wcfxs_dma_free(sc, &sc->sc_read);
	wcfxs_dma_free(sc, &sc->sc_write);

	/* Reset PCI chip and registers */
	WCFXS_WRITE1(WC_CNTL, 0x0e);

	wcfxs_release(sc);

	if (sc->sc_iosize)
	    bus_space_unmap(sc->sc_iot, sc->sc_ioh, sc->sc_iosize);

	if (card_configuration) {
	    free(card_configuration, M_TDM);
	    card_configuration = NULL;
	}

#ifdef _LKM
	malloc_type_detach(M_TDM);  /* XXX this  belongs elsewhere */
#endif
	aprint_naive("FXS detach!\n");

	return (0);
}

/* wcfxs_intr: interrupt service routine */
static int
wcfxs_intr(void *arg)
{
	struct wcfxs_softc *sc = arg;
	unsigned char ints;
	int x;

	ints = WCFXS_READ1(WC_INTSTAT);
	DPRINTF(2,("wcfxs_intr: 0x%x\n",ints));
	WCFXS_WRITE1(WC_INTSTAT, ints);

	if (!ints)
		return 0;

	if (ints & 0x10) {
		/* Stop DMA, wait for watchdog */
		printf("TDM PCI Master abort\n");
		wcfxs_stop_dma(sc);
		return 1;
	}

	if (ints & 0x20) {
		printf("PCI Target abort\n");
		return 1;
	}

	for (x = 0; x < 4; x++) {
	    if ((x < sc->sc_cards) && (sc->sc_cardmap & (1 << x)) &&
		(sc->sc_modtype[x] == MOD_TYPE_FXS)) {
		    if (sc->sc_mod.fxs.lasttxhook[x] == 0x4) {
			/* RINGing, prepare for OHT */
			sc->sc_mod.fxs.ohttimer[x] = OHT_TIMER << 3;
			/* OHT mode when idle */
			sc->sc_mod.fxs.idletxhookstate[x] = 0x2;
		    } else {
			if (sc->sc_mod.fxs.ohttimer[x]) {
			    sc->sc_mod.fxs.ohttimer[x] -= ZT_CHUNKSIZE;
			    if (!sc->sc_mod.fxs.ohttimer[x]) {
				/* switch to active */
				sc->sc_mod.fxs.idletxhookstate[x] = 1;
				if (sc->sc_mod.fxs.lasttxhook[x] == 2) {
				    /* Apply the change if appropriate */
				    sc->sc_mod.fxs.lasttxhook[x] = 1;
				    wcfxs_setreg(sc, x, 64,
					sc->sc_mod.fxs.lasttxhook[x]);
				}
			    }
			}
		    }
	    }
	}

	if (ints & 0x0f){
	    sc->sc_intcount++;
	    x = sc->sc_intcount % 4;
	    if ((x < sc->sc_cards) && (sc->sc_cardmap & (1 << x))) {
		if (sc->sc_modtype[x] == MOD_TYPE_FXS) {
		    wcfxs_proslic_check_hook(sc, x);
		    if (!(sc->sc_intcount & 0xfc))
			wcfxs_proslic_recheck_sanity(sc, x);
		} else if (sc->sc_modtype[x] == MOD_TYPE_FXO) {
		    wcfxs_voicedaa_check_hook(sc, x);
		}
	    }
	    if (!(sc->sc_intcount % 10000)) {
		/* Accept an alarm once per 10 seconds */
		for (x = 0; x < 4; x++)
		    if (sc->sc_modtype[x] == MOD_TYPE_FXS)
			if (sc->sc_mod.fxs.palarms[x])
			    sc->sc_mod.fxs.palarms[x]--;
	    }
	    wcfxs_receiveprep(sc, ints);
	    wcfxs_transmitprep(sc, ints);
	}

	return 1;
}

static int wcfxs_ioctl(struct zt_chan *chan, unsigned int cmd, caddr_t data)
{
	struct wcfxs_stats stats;
	struct wcfxs_regs regs;
	struct wcfxs_regop regop;
	struct wcfxs_softc *sc = chan->pvt;
	int x;
	switch (cmd) {
	case ZT_ONHOOKTRANSFER:
		if (sc->sc_modtype[chan->chanpos - 1] != MOD_TYPE_FXS)
			return EINVAL;
		bcopy((int *)data, &x, sizeof(int));
		sc->sc_mod.fxs.ohttimer[chan->chanpos - 1] = x << 3;
		sc->sc_mod.fxs.idletxhookstate[chan->chanpos - 1] = 0x2;	/* OHT mode when idle */
		if (sc->sc_mod.fxs.lasttxhook[chan->chanpos - 1] == 0x1) {
				/* Apply the change if appropriate */
				sc->sc_mod.fxs.lasttxhook[chan->chanpos - 1] = 0x2;
				wcfxs_setreg(sc, chan->chanpos - 1, 64, sc->sc_mod.fxs.lasttxhook[chan->chanpos - 1]);
		}
		break;

	case WCFXS_GET_STATS:
		if (sc->sc_modtype[chan->chanpos - 1] != MOD_TYPE_FXS)
			return EINVAL;
		stats.tipvolt = wcfxs_getreg(sc, chan->chanpos - 1, 80) * -376;
		stats.ringvolt = wcfxs_getreg(sc, chan->chanpos - 1, 81) * -376;
		stats.batvolt = wcfxs_getreg(sc, chan->chanpos - 1, 82) * -376;
		bcopy(&stats, (struct wcfxs_stats *)data, sizeof(stats));
		break;
	case WCFXS_GET_REGS:
		if (sc->sc_modtype == MOD_TYPE_FXS) {
			for (x=0;x<NUM_INDIRECT_REGS;x++)
				regs.indirect[x] = wcfxs_proslic_getreg_indirect(sc, chan->chanpos -1, x);
			for (x=0;x<NUM_REGS;x++)
				regs.direct[x] = wcfxs_getreg(sc, chan->chanpos - 1, x);
		} else {
			memset(&regs, 0, sizeof(regs));
			for (x=0;x<NUM_FXO_REGS;x++)
				regs.direct[x] = wcfxs_getreg(sc, chan->chanpos - 1, x);
		}
		bcopy(&regs, (struct wcfxs_regs *)data, sizeof(regs));
		break;
	case WCFXS_SET_REG:
		bcopy((struct wcfxs_regop *)data, &regop, sizeof(regop));
		if (regop.indirect) {
			if (sc->sc_modtype[chan->chanpos - 1] != MOD_TYPE_FXS)
				return EINVAL;
			printf("Setting indirect %d to 0x%04x on %d\n", regop.reg, regop.val, chan->chanpos);
			wcfxs_proslic_setreg_indirect(sc, chan->chanpos - 1, regop.reg, regop.val);
		} else {
			regop.val &= 0xff;
			printf("Setting direct %d to %04x on %d\n", regop.reg, regop.val, chan->chanpos);
			wcfxs_setreg(sc, chan->chanpos - 1, regop.reg, regop.val);
		}
		break;
	default:
		return ENOTTY;
	}
	return 0;

}

int wcfxs_open(struct zt_chan *chan)
{
	struct wcfxs_softc *sc = chan->pvt;
	if (!(sc->sc_cardmap & (1 << (chan->chanpos - 1))))
		return ENODEV;
	if (sc->sc_dead)
		return ENODEV;
	sc->sc_usecount++;
	return 0;
}

static int wcfxs_watchdog(struct zt_span *span, int event)
{
	printf("TDM: Restarting DMA\n");
	wcfxs_restart_dma(span->pvt);
	return 0;
}

static int wcfxs_close(struct zt_chan *chan)
{
	struct wcfxs_softc *sc = chan->pvt;
	int x;
	sc->sc_usecount--;
	for (x=0;x<sc->sc_cards;x++)
		sc->sc_mod.fxs.idletxhookstate[x] = 1;
	/* Actualy it is useless in BSD version but let it remains */
	/* If we're dead, release us now */
	if (!sc->sc_usecount && sc->sc_dead) 
		wcfxs_release(sc);
	return 0;
}

static void wcfxs_release(struct wcfxs_softc *sc)
{
	zt_unregister(&sc->sc_span);
}

static int wcfxs_hooksig(struct zt_chan *chan, zt_txsig_t txsig)
{
	struct wcfxs_softc *sc = chan->pvt;
	int reg=0;

	if (sc->sc_modtype[chan->chanpos - 1] == MOD_TYPE_FXO) {
		/* XXX Enable hooksig for FXO XXX */
		switch(txsig) {
		case ZT_TXSIG_START:
		case ZT_TXSIG_OFFHOOK:
			sc->sc_mod.fxo.offhook[chan->chanpos - 1] = 1;
			wcfxs_setreg(sc, chan->chanpos - 1, 5, 0x9);
			break;
		case ZT_TXSIG_ONHOOK:
			sc->sc_mod.fxo.offhook[chan->chanpos - 1] = 0;
			wcfxs_setreg(sc, chan->chanpos - 1, 5, 0x8);
 			break;
		default:
			printf("wcfxo: Can't set tx state to %d\n", txsig);
		}
	} else {
		switch(txsig) {
		case ZT_TXSIG_ONHOOK:
			switch(chan->sig) {
			case ZT_SIG_EM:
			case ZT_SIG_FXOKS:
			case ZT_SIG_FXOLS:
				sc->sc_mod.fxs.lasttxhook[chan->chanpos-1] = sc->sc_mod.fxs.idletxhookstate[chan->chanpos-1];
				break;
			case ZT_SIG_FXOGS:
				sc->sc_mod.fxs.lasttxhook[chan->chanpos-1] = 3;
				break;
			}
			break;
		case ZT_TXSIG_OFFHOOK:
			switch(chan->sig) {
				case ZT_SIG_EM:
					sc->sc_mod.fxs.lasttxhook[chan->chanpos-1] = 5;
					break;
				default:
					sc->sc_mod.fxs.lasttxhook[chan->chanpos-1] = sc->sc_mod.fxs.idletxhookstate[chan->chanpos-1];
				break;
			}
			break;
		case ZT_TXSIG_START:
			sc->sc_mod.fxs.lasttxhook[chan->chanpos-1] = 4;
			break;
		case ZT_TXSIG_KEWL:
			sc->sc_mod.fxs.lasttxhook[chan->chanpos-1] = 0;
			break;
		default:
			printf("wcfxs: Can't set tx state to %d\n", txsig);
		}
		DPRINTF(1,("Setting FXS hook state to %d (%02x)\n", txsig, reg));
 
#if 1
		wcfxs_setreg(sc, chan->chanpos - 1, 64, sc->sc_mod.fxs.lasttxhook[chan->chanpos-1]);
#endif
	}
	return 0;
}

#if 0

const struct cdevsw wcfxs_cdevsw = {
	wcfxs_open,
	wcfxs_close,
	wcfxs_read,
	wcfxs_write,
	wcfxs_ioctl,
	nostop,
	notty,
	nopoll,
	nommap,
	nokqfilter,
	0,
};

MOD_DEV("wcfxs", "wcfxs", NULL, -1, &wcfxs_cdevsw, -1);

int
mod_wcfxs_lkmentry(struct lkm_table *lkmtp, int cmd, int ver) {
	DISPATCH(lkmtp, cmd, ver, wcfxs_handle, wcfxs_handle, wcfxs_handle);
}
#endif
