/*
 * Dummy Zaptel Driver for Zapata Telephony interface
 *
 * Required: usb-uhci module and kernel > 2.4.4 OR kernel > 2.6.0
 *
 * Written by Robert Pleh <robert.pleh@hermes.si>
 * 2.6 version by Tony Hoyle
 * Unified by Mark Spencer <markster@digium.com>
 *
 * Copyright (C) 2002, Hermes Softlab
 * Copyright (C) 2004, Digium, Inc.
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <sys/cdefs.h>

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/uio.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#ifdef __FreeBSD__
#include <sys/module.h>
#endif
#ifdef __NetBSD__
#include <sys/lkm.h>
#endif

#include <sys/conf.h>
#include <sys/errno.h>

#include <zaptel.h>

#include "ztdummy.h"

MALLOC_DEFINE(M_ZTD, "ztdummy", "ztdummy interface data structures");

#ifdef __FreeBSD__
MODULE_VERSION(ztdummy, 1);

static struct callout_handle ztdummy_timer_handle = CALLOUT_HANDLE_INITIALIZER(&ztdummy_timer_handle);
#endif
#ifdef __NetBSD__
static struct callout ztdummy_callout = CALLOUT_INITIALIZER;
#endif

static struct ztdummy *ztd;

static int debug = 1;

static __inline void ztdummy_timer(void* arg )
{
	int i,j;

	j = max(1, 1000/hz);
	
	for(i=0; i < j; i++) {
		zt_receive(&ztd->span);
		zt_transmit(&ztd->span);
	}
	
#ifdef __FreeBSD__
	ztdummy_timer_handle = timeout(ztdummy_timer, NULL, hz/1000);
#endif
#ifdef __NetBSD__
	callout_reset(&ztdummy_callout, hz/1000 ? hz/1000 : 1,
		ztdummy_timer, NULL);
#endif

}

static int ztdummy_initialize(struct ztdummy *ztd)
{
	/* Zapata stuff */
	sprintf(ztd->span.name, "ZTDUMMY/1");
	sprintf(ztd->span.desc, "%s %d", ztd->span.name, 1);
	sprintf(ztd->chan.name, "ZTDUMMY/%d/%d", 1, 0);
	ztd->chan.chanpos = 1;
	ztd->span.chans = &ztd->chan;
	ztd->span.channels = 0;		/* no channels on our span */
	ztd->span.deflaw = ZT_LAW_MULAW;
	ztd->span.pvt = ztd;
	ztd->chan.pvt = ztd;
	if (zt_register(&ztd->span, 0)) {
		return -1;
	}
	return 0;
}

static int ztdummy_attach(void )
{

#ifdef __NetBSD__
	malloc_type_attach(M_ZTD);
#endif
	ztd = malloc(sizeof(struct ztdummy), M_ZTD, M_NOWAIT);
	if (ztd == NULL) {
		printf("ztdummy: Unable to allocate memory\n");
		return -ENOMEM;
	}

	memset(ztd, 0x0, sizeof(struct ztdummy));

	if (ztdummy_initialize(ztd)) {
		printf("ztdummy: Unable to intialize zaptel driver\n");
		free(ztd, M_ZTD);
		return -ENODEV;
	}

#ifdef __FreeBSD__
	ztdummy_timer_handle = timeout(ztdummy_timer, NULL, hz/1000);
#endif
#ifdef __NetBSD__
	callout_reset(&ztdummy_callout, hz/1000 ? hz/1000 : 1,
		ztdummy_timer, NULL);
#endif
    
	if (debug)
		printf("ztdummy: init() finished\n");
	return 0;
}


static void cleanup_module(void)
{
#ifdef __FreeBSD__
	untimeout(ztdummy_timer, NULL, ztdummy_timer_handle);
#endif
#ifdef __NetBSD__
	callout_stop(&ztdummy_callout);
#endif
#ifdef __NetBSD__
	malloc_type_detach(M_ZTD);
#endif
	zt_unregister(&ztd->span);
	free(ztd, M_ZTD);

	if (debug)
		printf("ztdummy: cleanup() finished\n");
}


#ifdef __FreeBSD__
static int ztdummy_modevent(module_t mod,int  type, void*  data)
#else
static int ztdummy_modevent(struct lkm_table *lkmtp, int type)
#endif
{

static int attached = 0;
int  ret;

 switch (type) {
 case MOD_LOAD:
	 if (attached)
		 return EEXIST;
	 
	 ret = ztdummy_attach();
	 if (ret)
		 return (ret);
	 attached = 1;
	 printf("ztdummy: loaded\n");
	 if (hz < 1000) {
		 printf ("ztdummy: WARNING Ticker rate only %d. Timer will not work well!!\nRecompile kernel with \"options HZ=1000\" \n", hz);
	 }
	 
	 break;

 case MOD_UNLOAD:
	 attached = 0;
	 cleanup_module ();
	 
	 printf("ztdummy: unloaded\n");
	 break;
	 
#ifdef __FreeBSD__
 case MOD_SHUTDOWN:
#endif
 default:
	 return EOPNOTSUPP;
 }
 return 0;
}


#ifdef __FreeBSD__
MODULE_DEPEND(ztdummy, zaptel, 1, 1, 1);
DEV_MODULE(ztdummy, ztdummy_modevent, NULL);
#endif
#ifdef __NetBSD__
/*MOD_DEV("ztdummy", "ztdummy", NULL, -1, NULL, -1)*/
MOD_MISC("ztdummy");

int
mod_ztdummy_lkmentry(struct lkm_table *lkmtp, int cmd, int ver)
{
	DISPATCH(lkmtp, cmd, ver, ztdummy_modevent, ztdummy_modevent,
			lkm_nofunc)
}
#endif
